__all__ = ["Results"]


import pickle
import warnings
import os
import glob
import copy

from astropy.io import fits
import astropy
import batman
import corner
import dynesty
import matplotlib.pyplot as plt
import numpy as np
import numpy.polynomial.polynomial as poly
import pandas as pd

from .io import load_detrended_lightcurve
from .utils import weighted_percentile, bin_data
from .constants import lcit, scit


class _Lightcurve:
    def __init__(self, key_values):
        self._keys = []

        for k, v in key_values.items():
            self._keys.append(k)
            setattr(self, k, copy.copy(v))

        required_keys = ["time", "flux", "error", "cadno", "quarter"]

        for k in required_keys:
            if k not in self._keys:
                raise ValueError("Key {0} must be provided".format(k))

    def keys(self):
        return self._keys

    def plot(self, quarter=None):
        if quarter is None:
            quarter = self.quarter

        plt.figure(figsize=(20, 4))
        plt.plot(
            self.time[self.quarter == quarter],
            self.flux[self.quarter == quarter],
            "k.",
            ms=1,
        )
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        plt.xlabel("Time [BKJD]", fontsize=24)
        plt.ylabel("Flux", fontsize=24)
        if np.any(self.quarter == quarter):
            plt.xlim(
                self.time[self.quarter == quarter].min(),
                self.time[self.quarter == quarter].max(),
            )
        else:
            warnings.warn("Attempting to plot quarter with no data")
        plt.show()

        return None


class _TransitTimes:
    def __init__(self, key_values):
        self._keys = []

        for k, v in key_values.items():
            self._keys.append(k.lower())
            setattr(self, k.lower(), np.array(v))

        required_keys = ["index", "ttime", "model", "out_prob", "out_flag"]

        self.index = self.index.astype(int)
        self.out_flag = self.out_flag.astype(bool)

        # calculate least squares period, epoch, and omc
        self.t0, self.period = poly.polyfit(self.index, self.model, 1)
        self.omc_ttime = self.ttime - poly.polyval(self.index, [self.t0, self.period])
        self.omc_model = self.model - poly.polyval(self.index, [self.t0, self.period])

        self._keys += ["t0", "period", "omc_ttime", "omc_model"]

    def keys(self):
        return self._keys


class _MultiTransitTimes:
    def __init__(self, _TTlist):
        self._keys = ["npl"] + _TTlist[0].keys()
        self.npl = len(_TTlist)

        for k in self._keys[1:]:
            v = [None] * self.npl
            for n, _TT in enumerate(_TTlist):
                v[n] = getattr(_TT, k)
            setattr(self, k, copy.copy(v))

    def keys(self):
        return self._keys

    def linear_ephemeris(self, n):
        return self.t0[n] + self.period[n] * self.index[n]


class _Posteriors:
    def __init__(self, key_values):
        self._keys = ["samples", "ln_like", "ln_wt", "ln_z"]

        df = pd.DataFrame(key_values)

        self.ln_like = np.array(df["LN_LIKE"])
        self.ln_wt = np.array(df["LN_WT"])
        self.ln_z = np.array(df["LN_Z"])

        self.samples = df.drop(columns=["LN_LIKE", "LN_WT", "LN_Z"])

    def keys(self):
        return self._keys

    def nsamp(self):
        """
        Number of samples
        """
        return self.samples.shape[0]

    def neff(self):
        """
        Number of effective samples (Kish ESS); wrapper for dynesty.utils.get_neff_from_logwt
        """
        return dynesty.utils.get_neff_from_logwt(self.ln_wt)

    def weights(self):
        """
        Convenience function for getting normalized weights
        """
        wt = np.exp(self.ln_wt - self.ln_z[-1])
        return wt / np.sum(wt)

    def summary(self):
        if hasattr(self, "_summary") == False:
            samples = self.samples
            weights = self.weights()
            npl = (samples.shape[1] - 2) // 5

            labels = list(self.samples.keys())
            _summary = pd.DataFrame(
                columns=[
                    "mean",
                    "stdev",
                    "median",
                    "CI_2.5",
                    "CI_16",
                    "CI_84",
                    "CI_97.5",
                ],
                index=labels,
            )

            for i, lab in enumerate(labels):
                avg = np.sum(weights * samples[lab])
                var = np.sum(weights * (samples[lab] - avg) ** 2)

                _summary["mean"][lab] = copy.copy(avg)
                _summary["stdev"][lab] = np.sqrt(var)
                _summary["median"][lab] = weighted_percentile(
                    samples[lab], 50, w=weights
                )
                _summary["CI_2.5"][lab] = weighted_percentile(
                    samples[lab], 2.5, w=weights
                )
                _summary["CI_16"][lab] = weighted_percentile(
                    samples[lab], 15.9, w=weights
                )
                _summary["CI_84"][lab] = weighted_percentile(
                    samples[lab], 84.1, w=weights
                )
                _summary["CI_97.5"][lab] = weighted_percentile(
                    samples[lab], 97.5, w=weights
                )

            self._summary = copy.copy(_summary)

        return self._summary


class Results:
    def __init__(self, target=None, data_dir=None):
        if target is not None:
            self.target = target

            if data_dir is not None:
                results_file = os.path.join(
                    data_dir, "{0}/{0}-results.fits".format(target)
                )

                with fits.open(results_file) as hduL:
                    if hduL[0].header["TARGET"] != self.target:
                        raise ValueError("specified target does not match FITS header")

                    self.npl = hduL[0].header["NPL"]

            else:
                self.npl = 0
                warnings.warn("Results object initiated without data")

        else:
            self.target = None
            warnings.warn("Results object initiated without target")

        # initialize extensions
        self.lightcurve = None
        self.transittimes = None
        self.posteriors = None

        self._keys = ["target", "npl", "lightcurve", "transittimes", "posteriors"]

        # read in lightcurves
        lightcurve_files = glob.glob(
            os.path.join(data_dir, "{0}/*filtered.fits".format(target))
        )
        for lcf in lightcurve_files:
            self.load_lightcurve(lcf)

        # read in results
        results_file = os.path.join(data_dir, "{0}/{0}-results.fits".format(target))
        self.load_posteriors(results_file)
        self.load_transittimes(results_file)

        # make samples container
        self._samples = [None] * self.npl

    def keys(self):
        return self._keys

    def methods(self):
        print(
            [
                "load_lightcurve",
                "load_posteriors",
                "load_transittimes",
                "plot_folded",
                "plot_lightcurve",
                "plot_omc",
                "plot_transit",
                "summary",
            ]
        )

    def load_lightcurve(self, f):
        _lightcurve = {}
        with fits.open(f) as hduL:
            _lightcurve["time"] = np.array(hduL["TIME"].data, dtype="float64")
            _lightcurve["flux"] = np.array(hduL["FLUX"].data, dtype="float64")
            _lightcurve["error"] = np.array(hduL["ERROR"].data, dtype="float64")
            _lightcurve["cadno"] = np.array(hduL["CADNO"].data, dtype="int")
            _lightcurve["quarter"] = np.array(hduL["QUARTER"].data, dtype="int")

        if self.lightcurve is not None:
            _lightcurve["time"] = np.hstack([self.lightcurve.time, _lightcurve["time"]])
            _lightcurve["flux"] = np.hstack([self.lightcurve.flux, _lightcurve["flux"]])
            _lightcurve["error"] = np.hstack(
                [self.lightcurve.error, _lightcurve["error"]]
            )
            _lightcurve["cadno"] = np.hstack(
                [self.lightcurve.cadno, _lightcurve["cadno"]]
            )
            _lightcurve["quarter"] = np.hstack(
                [self.lightcurve.quarter, _lightcurve["quarter"]]
            )

        self.lightcurve = _Lightcurve(_lightcurve)

    def load_posteriors(self, f):
        with fits.open(f) as hduL:
            data = hduL["SAMPLES"].data
            keys = data.names

            _posteriors = []
            for k in keys:
                _posteriors.append(data[k])

            self.posteriors = _Posteriors(
                pd.DataFrame(np.array(_posteriors).T, columns=keys)
            )

    def load_transittimes(self, f):
        _transittimes = [None] * self.npl

        with fits.open(f) as hduL:
            for n in range(self.npl):
                data = hduL["TTIMES_{0}".format(str(n).zfill(2))].data
                #data = hduL[n+2].data
                keys = data.names

                _tts = []
                for k in keys:
                    _tts.append(data[k])

                _transittimes[n] = _TransitTimes(
                    pd.DataFrame(np.array(_tts).T, columns=keys)
                )

        self.transittimes = _MultiTransitTimes(_transittimes)

    def plot_lightcurve(self, quarter=None):
        self.lightcurve.plot(quarter)

    def plot_omc(self, show_outliers=False):
        fig, ax = plt.subplots(self.npl, figsize=(12, 3 * self.npl))
        if self.npl == 1:
            ax = [ax]
        for i in range(self.npl):
            tts = self.transittimes.ttime[i]
            out = self.transittimes.out_flag[i]
            omc_ttime = self.transittimes.omc_ttime[i]
            omc_model = self.transittimes.omc_model[i]

            if show_outliers:
                ax[i].scatter(
                    tts,
                    omc_ttime * 24 * 60,
                    c=self.transittimes.out_prob[i],
                    cmap="viridis",
                )
                ax[i].plot(tts[out], omc_ttime[out] * 24 * 60, "rx")
                ax[i].plot(tts, omc_model * 24 * 60, "k")
            else:
                ax[i].plot(tts[~out], omc_ttime[~out] * 24 * 60, "o", c="lightgrey")
                ax[i].plot(
                    tts[~out], omc_model[~out] * 24 * 60, c="C{0}".format(i), lw=3
                )
            ax[i].set_ylabel("O-C [min]", fontsize=20)

        ax[self.npl - 1].set_xlabel("Time [BJKD]", fontsize=20)
        plt.show()

    def plot_transit(self, n, index):
        loc = self.transittimes.index[n] == index
        tc = float(self.transittimes.model[n][loc])
        T14 = self.posteriors.summary()["median"]["DUR14_{0}".format(n)]
        use = np.abs(self.lightcurve.time - tc) / T14 < 1.5

        theta = self._batman_theta(n)

        t_obs = self.lightcurve.time[use]
        f_obs = (self.lightcurve.flux[use] - 1.0) * 1000
        t_mod = np.arange(t_obs.min(), t_obs.max(), scit)
        f_mod = (batman.TransitModel(theta, t_mod - tc).light_curve(theta) - 1.0) * 1000

        plt.figure(figsize=(12, 4))
        plt.plot(t_mod, f_mod, color="C{0}".format(n), lw=2)
        plt.plot(t_obs, f_obs, "o", color="lightgrey")
        plt.xlim(tc - 1.55 * T14, tc + 1.55 * T14)
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        plt.xlabel("Time [BKJD]", fontsize=24)
        plt.ylabel("Flux [ppt]", fontsize=24)
        plt.text(
            t_obs.max(),
            f_obs.max(),
            "PERIOD = {0}".format(np.round(theta.per, 2)),
            fontsize=20,
            ha="right",
            va="top",
        )
        plt.show()

    def plot_folded(self, n, max_pts=1000):
        # grab posteriors
        time = self.lightcurve.time
        flux = self.lightcurve.flux
        tts = self.transittimes.model[n]
        T14 = self.posteriors.summary()["median"]["DUR14_{0}".format(n)]

        # fold photometry
        t_folded = []
        f_folded = []

        for t0 in tts:
            use = np.abs(time - t0) / T14 < 1.5
            t_folded.append(time[use] - t0)
            f_folded.append(flux[use])

        t_folded = np.hstack(t_folded)
        f_folded = np.hstack(f_folded)

        t_binned, f_binned = bin_data(t_folded, f_folded, T14 / 11)

        theta = self._batman_theta(n)

        t_mod_sc = np.arange(t_folded.min(), t_folded.max(), scit)
        t_mod_lc = np.arange(t_folded.min(), t_folded.max(), lcit / 29)

        f_mod_sc = batman.TransitModel(theta, t_mod_sc).light_curve(theta)
        f_mod_lc = batman.TransitModel(
            theta, t_mod_lc, supersample_factor=29, exp_time=lcit
        ).light_curve(theta)

        f_pred = batman.TransitModel(
            theta, t_folded, supersample_factor=29, exp_time=lcit
        ).light_curve(theta)

        residuals = f_folded - f_pred
        _, res_binned = bin_data(t_folded, residuals, T14 / 11)

        # estimate contact points to assess differential scatter
        s = self.posteriors.summary()

        T14 = s.at["DUR14_{0}".format(n), "median"]
        ror = s.at["ROR_{0}".format(n), "median"]
        b = s.at["IMPACT_{0}".format(n), "median"]

        T23_over_T14 = np.sqrt(((1 - ror) ** 2 - b**2) / ((1 + ror) ** 2 - b**2))

        T23 = T14 * T23_over_T14
        tau = (T14 - T23) / 2

        t_ = t_folded
        f_ = f_folded
        r_ = residuals

        rms = {}
        rms["before"] = np.round(
            astropy.stats.mad_std(r_[(t_ > -T14 - tau) * (t_ < -T14)]) * 1e6, 1
        )
        rms["ingress"] = np.round(
            astropy.stats.mad_std(r_[(t_ > -T14 / 2) * (t_ < -T23 / 2)]) * 1e6, 1
        )
        rms["mid"] = np.round(
            astropy.stats.mad_std(r_[(t_ > -tau / 2) * (t_ < tau / 2)]) * 1e6, 1
        )
        rms["egress"] = np.round(
            astropy.stats.mad_std(r_[(t_ > T23 / 2) * (t_ < T14 / 2)]) * 1e6, 1
        )
        rms["after"] = np.round(
            astropy.stats.mad_std(r_[(t_ > T14) * (t_ < T14 + tau)]) * 1e6, 1
        )

        # select a subset of points to plot
        inds = np.arange(len(t_folded), dtype="int")
        inds = np.random.choice(inds, size=np.min([max_pts, len(inds)]), replace=False)

        # make the plot
        fig = plt.figure(figsize=(12, 6))

        ax = [
            plt.subplot2grid(shape=(3, 1), loc=(0, 0), rowspan=2, colspan=1),
            plt.subplot2grid(shape=(3, 1), loc=(2, 0), rowspan=1, colspan=1),
        ]

        ax[0].plot(t_folded[inds] * 24, f_folded[inds], "o", color="lightgrey")
        ax[0].plot(t_binned * 24, f_binned, "s", color="w", mec="k", ms=10)
        ax[0].plot(t_mod_sc * 24, f_mod_sc, color="C{0}".format(n), lw=3)
        ax[0].plot(t_mod_lc * 24, f_mod_lc, "k--", lw=2)
        ax[0].set_xlim(t_folded.min() * 24, t_folded.max() * 24)
        ax[0].set_ylabel("Flux", fontsize=24)

        yloc = 0.5 * (f_.max() + f_.min())

        ax[0].text(-T14 / 2, yloc, rms["before"], color="C{0}".format(n))

        ax[1].plot(t_folded[inds] * 24, residuals[inds], ".", color="lightgrey")
        ax[1].plot(t_binned * 24, res_binned, "s", color="w", mec="k", ms=10)
        ax[1].axhline(0, color="k", ls="--", lw=2)
        ax[1].set_xlim(t_folded.min() * 24, t_folded.max() * 24)
        ax[1].set_xlabel("Time from mid-transit [hrs]", fontsize=24)
        ax[1].text(
            0.02,
            0.8,
            "RMS = {0}".format(int(np.std(residuals) * 1e6)),
            transform=ax[1].transAxes,
            fontsize=11,
        )

        plt.show()

    def summary(self):
        return self.posteriors.summary()

    def samples(self, n):
        if self._samples[n] is None:
            # raw posteriors
            C0 = self.posteriors.samples["C0_{0}".format(n)]
            C1 = self.posteriors.samples["C1_{0}".format(n)]
            r = self.posteriors.samples["ROR_{0}".format(n)]
            b = self.posteriors.samples["IMPACT_{0}".format(n)]
            T14 = self.posteriors.samples["DUR14_{0}".format(n)]
            q1 = self.posteriors.samples["LD_Q1"]
            q2 = self.posteriors.samples["LD_Q2"]
            ln_wt = self.posteriors.ln_wt

            # least squares period and epoch
            centered_index = (
                self.transittimes.index[n] - self.transittimes.index[n][-1] // 2
            )

            LegX = centered_index / (self.transittimes.index[n][-1] / 2)
            Leg0 = np.ones_like(LegX)
            ephem = self.transittimes.model[n] + np.outer(C0, Leg0) + np.outer(C1, LegX)
            t0, P = poly.polyfit(self.transittimes.index[n], ephem.T, 1)

            # limb darkening (see Kipping 2013)
            u1 = 2 * np.sqrt(q1) * q2
            u2 = np.sqrt(q1) * (1 - 2 * q2)

            # build dataframe
            data = np.vstack([P, t0, r, b, T14, u1, u2, ln_wt]).T
            labels = "PERIOD T0 ROR IMPACT DUR14 LD_U1 LD_U2 LN_WT".split()
            self._samples[n] = pd.DataFrame(data, columns=labels)

        return self._samples[n]

    def _batman_theta(self, n):
        theta = batman.TransitParams()
        theta.per = self.transittimes.period[n]
        theta.t0 = 0.0
        theta.rp = self.posteriors.summary()["median"]["ROR_{0}".format(n)]
        theta.b = self.posteriors.summary()["median"]["IMPACT_{0}".format(n)]
        theta.T14 = self.posteriors.summary()["median"]["DUR14_{0}".format(n)]

        q1 = self.posteriors.summary()["median"]["LD_Q1"]
        q2 = self.posteriors.summary()["median"]["LD_Q2"]
        theta.u = [2 * np.sqrt(q1) * q2, np.sqrt(q1) * (1 - 2 * q2)]
        theta.limb_dark = "quadratic"

        return theta



__all__ = ['BaseAlg']

import numpy as np
from alderaan.constants import kepler_lcit, kepler_scit
from alderaan.litecurve import LiteCurve
from alderaan.planet import Planet


class BaseAlg():
    """
    Base Algorithm class for modules which work on a litecurve + planets
     * Detrend
     * Transit Model
     * TTV Model
     * Quality Control
    """
    def __init__(self, litecurve, planets):
        # check inputs
        if not isinstance(litecurve, LiteCurve):
            raise TypeError("litecurve must be a LiteCurve")
        if not isinstance(planets, list):
            raise TypeError("planets must be a list of Planets")
        for p in planets:
            if not isinstance(p, Planet):
                raise TypeError("planets must be a list of Planets")

        # set attributes
        self.litecurve = litecurve
        self.npl = len(planets)
        self.planets = planets

        self.periods = np.zeros(self.npl, dtype=float)
        self.depths = np.zeros(self.npl, dtype=float)
        self.durs = np.zeros(self.npl, dtype=float)
        
        for n, p in enumerate(planets):
            self.periods[n] = p.period
            self.depths[n] = p.depth
            self.durs[n] = p.duration

        # define lookup
        self._init_obsmode_tracking()


    def _init_obsmode_tracking(self):
        self.transit_obsmode = self.get_transit_obsmode()
        self.unique_obsmodes = np.unique(np.hstack(self.transit_obsmode))

        self._define_exptime_lookup()
        self._define_supersample_lookup()
        self._define_exptime_integration_offset_lookup()


    def _define_exptime_lookup(self):
        self._exptime_lookup = {
            'long cadence': kepler_lcit,
            'short cadence': kepler_scit,
        }


    def _obsmode_to_exptime(self, obsmode):       
        return self._exptime_lookup[obsmode]
    

    def _compute_supersample(self, obsmode):    
        # ingress/egress timescale estimate following Winn 2010
        tau12 = (13 / 24) * (np.array(self.periods) / 365.25) ** (1 / 3) * np.sqrt(np.array(self.depths)) 

        # sigma so binning error is < 0.1% of photometric uncertainty
        sigma = np.nanmean(self.litecurve.error / self.litecurve.flux) * 0.04

        # supersample factor following Kipping 2010
        exptime = self._obsmode_to_exptime(obsmode)
        supersample = np.array(np.ceil(np.sqrt((self.depths / tau12) * (exptime / 8 / sigma))), dtype=int)
        supersample = np.max([supersample + (supersample % 2 + 1)])

        return supersample
    

    def _define_supersample_lookup(self):
        self._supersample_lookup = {}
        for om in np.unique(np.hstack(self.transit_obsmode)):
            self._supersample_lookup[om] = self._compute_supersample(om)
        

    def _obsmode_to_supersample(self, obsmode):
        return self._supersample_lookup[obsmode]


    def _compute_exptime_integration_offset(self, obsmode):
        exptime = self._obsmode_to_exptime(obsmode)
        supersample = self._obsmode_to_supersample(obsmode)
        return np.linspace(-exptime/2, exptime/2, supersample)
    

    def _define_exptime_integration_offset_lookup(self):
        self._exptime_integration_offset_lookup = {}
        for om in np.unique(np.hstack(self.transit_obsmode)):
            self._exptime_integration_offset_lookup[om] = (
                self._compute_exptime_integration_offset(om)
            )
            

    def _obsmode_to_exptime_integration_offset(self, obsmode):
        return self._exptime_integration_offset_lookup[obsmode]
    
            
    def get_transit_obsmode(self):
        """
        Determine the observing mode at each transit time
        Returns a length num_planets list, each entry is a list of obsmode str
        """
        lc = self.litecurve
        obsmode = [None]*self.npl

        for n, p in enumerate(self.planets):
            obsmode[n] = []
            for tc in p.ephemeris.ttime:
                obsmode[n].append(lc.obsmode[np.argmin(np.abs(lc.time-tc))])
            obsmode[n] = np.array(obsmode[n])

        return obsmode


    def make_transit_mask(self, rel_size=None, abs_size=None, mask_type='standard'):
        """
        Arguments
            rel_size : full width of masked region in units of transit duration
            abs_size : full width of masked region in units of time
            mask_type : type of transit mask to return (default = 'standard')
        
        Several output mask_type are supported
         * standard : bool, shape (n_planet, n_cadence), True near transits
         * condensed : bool, shape (n_cadence), True near transit
         * count : int, shape (n_cadence), value is number of planets near transit
         * overlap : bool, shanpe (n_cadnece), True if multiple planets are near transit
        """
        if (rel_size is None) & (abs_size is None):
            raise ValueError("either rel_size or abs_size must be provided")
        
        if rel_size is None:
            rel_size = 0.
        if abs_size is None:
            abs_size = 0.

        mask = np.zeros((self.npl,len(self.litecurve.time)), dtype=int)

        for n, p in enumerate(self.planets):
            half_width = 0.5*np.max([rel_size*p.duration, abs_size])
            for tc in p.ephemeris.ttime:
                mask[n] += np.abs(self.litecurve.time - tc) < half_width

        if mask_type == 'standard':
            mask = mask.astype(bool)
        elif mask_type == 'condensed':
            mask = np.sum(mask, axis=0).astype(bool)
        elif mask_type == 'count':
            mask = np.sum(mask, axis=0)
        elif mask_type == 'overlap':
            mask = np.sum(mask, axis=0) > 0
        else:
            raise ValueError(f"mask_type f{mask_type} not supported")

        return mask.astype(bool)


    def identify_overlapping_transits(self, rtol=None, atol=None):
        """
        Identify where transits overlap based on separation of transit midpoints

        Arguments
            rtol (float) : relative tolerance, in units of transit durations
            atol (float) : absolute tolerance, in units of hours
        """
        if (rtol is None) and (atol is None):
            raise ValueError("at least one of rtol or atol must be provided")

        overlap = [None]*self.npl

        for i in range(self.npl):
            overlap[i] = np.zeros(len(self.planets[i].ephemeris.ttime), dtype=bool)
            ttime_i = self.planets[i].ephemeris.ttime
            
            for j in range(i+1,self.npl):
                for tc_j in self.planets[j].ephemeris.ttime:
                    if rtol is not None:
                        overlap[i] |= np.abs(ttime_i - tc_j) < rtol * 0.5 * (self.durs[i] + self.durs[j])
                    if atol is not None:
                        overlap[i] |= np.abs(ttime_i - tc_j) < atol

        return overlap


__all__ = [
    "remove_bad_cadences",
    "estimate_oscillation_period",
    "identify_gaps",
    "flatten_litecurve",
    "flatten_with_gp",
    "stitch",
    "filter_ringing",
]


from copy import deepcopy
import warnings

import aesara_theano_fallback.tensor as T
import astropy
from astropy.timeseries import LombScargle
from celerite2.backprop import LinAlgError
from celerite2.theano import GaussianProcess
from celerite2.theano import terms as GPterms
import lightkurve as lk
import numpy as np
import pymc3 as pm
import pymc3_ext as pmx
import scipy.signal as sig

from .astro import make_transit_mask
from .constants import *


def remove_bad_cadences(litecurve, planets, rel_masksize, min_masksize):
    """
    Docstring
    """
    # remove flagged cadences
    qmask = lk.KeplerQualityFlags.create_quality_mask(
        litecurve.quality, bitmask="default"
    )
    litecurve.remove_flagged_cadences(qmask)

    # make transit mask
    litecurve.mask = np.zeros(len(litecurve.time), dtype="bool")
    for n, p in enumerate(planets):
        masksize = np.max([min_masksize, rel_masksize * p.duration])
        litecurve.mask += make_transit_mask(litecurve.time, p.tts, masksize)

    litecurve.clip_outliers(
        kernel_size=13, sigma_upper=5, sigma_lower=5, mask=litecurve.mask
    )
    litecurve.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=1000, mask=None)

    return litecurve


def estimate_oscillation_period(litecurve, min_period):
    """
    Docstring
    """
    ls_estimate = LombScargle(litecurve.time, litecurve.flux)

    min_freq = 1 / (litecurve.time.max() - litecurve.time.min())
    max_freq = 1 / min_period

    xf, yf = ls_estimate.autopower(
        minimum_frequency=min_freq, maximum_frequency=max_freq
    )

    peak_freq = xf[np.argmax(yf)]
    peak_per = np.max([1.0 / peak_freq, 1.001 * min_period])

    return peak_per


def identify_gaps(lc, break_tolerance, jump_tolerance=5.0):
    """
    Find gaps (breaks in time) and jumps (sudden flux changes) in a LiteCurve

    Parameters
    ----------
        lc : alderaan.LiteCurve
            Litecurve to be analyzed
        break_tolerance : int
            number of cadences to be considered a (time) gap
        jump_tolerance : float
            sigma threshold for identifying (flux) jumps (i.e. cadence-to-cadence flux variation)

    Returns
    -------
        gaps : ndarray, dtype=int
            array of indexes corresponding to the locations of gaps in lc.flux
    """
    # 1D mask
    mask = np.sum(np.atleast_2d(lc.mask, 0) == 0)

    # identify time gaps
    breaks = lc.cadno[1:] - lc.cadno[:-1]
    breaks = np.pad(breaks, (1, 0), "constant", constant_values=(1, 0))
    break_locs = np.where(breaks > break_tolerance)[0]
    break_locs = np.pad(
        break_locs, (1, 1), "constant", constant_values=(0, len(breaks) + 1)
    )

    # identify flux jumps
    jumps = lc.flux[1:] - lc.flux[:-1]
    jumps = np.pad(jumps, (1, 0), "constant", constant_values=(0, 0))
    big_jump = (
        np.abs(jumps - np.median(jumps)) / astropy.stats.mad_std(jumps) > jump_tolerance
    )
    jump_locs = np.where(mask * big_jump)[0]

    gaps = np.sort(np.unique(np.hstack([break_locs, jump_locs])))

    # flag nearly-consecutive cadences identified as gaps
    bad = np.hstack([False, (gaps[1:] - gaps[:-1]) < break_tolerance])

    if bad[-1]:
        bad[-1] = False
        bad[-2] = True

    gaps = gaps[~bad]

    return gaps


def flatten_litecurve(
    litecurve, break_tolerance, min_period, nominal_period, verbose=False
):
    """
    Remove trends from a LiteCurve using celerite Gaussian Process
    Calls flatten_with_gp() using successively simpler models as fallbacks
    """
    try:
        litecurve = flatten_with_gp(
            litecurve,
            break_tolerance,
            min_period,
            nominal_period=nominal_period,
            kterm="RotationTerm",
            correct_ramp=True,
            verbose=verbose,
        )
    except LinAlgError:
        warnings.warn(
            "Initial detrending failed...attempting to refit without exponential ramp component"
        )

        try:
            litecurve = flatten_with_gp(
                litecurve,
                break_tolerance,
                min_period,
                nominal_period=nominal_period,
                kterm="RotationTerm",
                correct_ramp=False,
                verbose=verbose,
            )
        except LinAlgError:
            warnings.warn(
                "Detrending with RotationTerm failed...attempting to detrend with SHOTerm"
            )

            litecurve = detrend.flatten_with_gp(
                litecurve,
                break_tolerance,
                min_period,
                nominal_period=nominal_period,
                kterm="SHOTerm",
                correct_ramp=False,
                verbose=verbose,
            )

    return litecurve


def flatten_with_gp(
    lc,
    break_tolerance,
    min_period,
    nominal_period=None,
    kterm="RotationTerm",
    correct_ramp=True,
    return_trend=False,
    verbose=True,
):
    """
    Remove trends from a LiteCurve using celerite Gaussian processes

    Parameters
    ----------
        lc : alderaan.LiteCurve
            LiteCurve to be flattened
        break_tolerance : int
            number of cadences to be considered a gap in data
        min_period : float
            minimum allowed period of GP kernel
        nominal_period : float
            mean estimate of period to pass to GP kernel
        kterm : string
            must be either 'RotationTerm' (default) or 'SHOTerm'
        correct_ramp : bool
            True to include an exponential ramp in the model for each disjoint section of photometry
        return_trend : bool
            True to return the predicted GP trend

    Returns
    -------
        lc : alderaan.LiteCurve
            LiteCurve with lc.flux and lc.error flattened and normalized
    """
    # identify primary oscillation period
    if nominal_period is None:
        ls_estimate = LombScargle(lc.time, lc.flux)
        xf, yf = ls_estimate.autopower(
            minimum_frequency=1 / (lc.time.max() - lc.time.min()),
            maximum_frequency=1 / min_period,
        )

        peak_freq = xf[np.argmax(yf)]
        peak_per = np.max([1.0 / peak_freq, 1.001 * min_period])

    else:
        peak_per = np.copy(nominal_period)

    # find gaps/jumps in the data
    gaps = identify_gaps(lc, break_tolerance=break_tolerance)
    gaps[-1] -= 1
    nseg = len(gaps) - 1

    # make adjustments for masked transits
    inds_ = np.arange(len(lc.time), dtype="int")[~lc.mask]
    gaps_ = [np.sum(inds_ < g) for g in gaps]
    time_ = lc.time[~lc.mask]
    flux_ = lc.flux[~lc.mask]

    # break up data into segments bases on gaps/jumps
    seg = np.zeros(len(lc.time), dtype="int")
    seg_ = np.zeros(len(time_), dtype="int")

    for i in range(nseg):
        seg[gaps[i] : gaps[i + 1]] = i
        seg_[gaps_[i] : gaps_[i + 1]] = i

    # define the mean function (exponential ramp)
    if correct_ramp:

        def mean_fxn(_t, _s, flux0, ramp_amp, log_tau):
            mean = T.zeros(len(_t))

            for i in range(len(np.unique(_s))):
                t0 = _t[_s == i].min()
                mean += (
                    flux0[i]
                    * (1 + ramp_amp[i] * T.exp(-(_t - t0) / T.exp(log_tau[i])))
                    * (_s == i)
                )

            return mean

    else:

        def mean_fxn(_t, _s, flux0, ramp_amp=None, log_tau=None):
            mean = T.zeros(len(_t))

            for i in range(len(np.unique(_s))):
                mean += flux0[i] * (_s == i)

            return mean

    # here's the stellar rotation model
    with pm.Model() as trend_model:
        # set up the kernal
        log_sigma = pm.Normal("log_sigma", mu=np.log(np.std(flux_)), sd=5.0)
        logP_off = pm.Normal("logP", mu=np.log(peak_per - min_period), sd=2.0)
        log_Q0 = pm.Normal("log_Q0", mu=0.0, sd=5.0, testval=np.log(0.5))

        sigma = pm.Deterministic("sigma", T.exp(log_sigma))
        P = pm.Deterministic("P", min_period + T.exp(logP_off))

        if kterm == "RotationTerm":
            log_dQ = pm.Normal("log_dQ", mu=0.0, sd=5.0, testval=np.log(1e-3))
            mix = pm.Uniform("mix", lower=0, upper=1, testval=0.1)
            kernel = GPterms.RotationTerm(
                sigma=sigma, period=P, Q0=T.exp(log_Q0), dQ=T.exp(log_dQ), f=mix
            )

        elif kterm == "SHOTerm":
            kernel = GPterms.SHOTerm(sigma=sigma, w0=2 * pi / P, Q=0.5 + T.exp(log_Q0))

        else:
            raise ValueError("kterm must be 'RotationTerm' or 'SHOTerm'")

        # mean function is an exponential trend (per segment)
        approx_mean_flux = [np.mean(flux_[seg_ == i]) for i in range(nseg)]

        if correct_ramp:
            flux0 = pm.Normal(
                "flux0", mu=approx_mean_flux, sd=np.ones(nseg), shape=nseg
            )
            ramp_amp = pm.Normal("ramp_amp", mu=0, sd=np.std(lc.flux), shape=nseg)
            log_tau = pm.Normal("log_tau", mu=0, sd=5, shape=nseg)
            mean_ = pm.Deterministic(
                "mean_", mean_fxn(time_, seg_, flux0, ramp_amp, log_tau)
            )

        else:
            flux0 = pm.Normal(
                "flux0", mu=approx_mean_flux, sd=np.ones(nseg), shape=nseg
            )
            ramp_amp = None
            log_tau = None
            mean_ = pm.Deterministic("mean_", mean_fxn(time_, seg_, flux0))

        # variance
        log_yvar = pm.Normal(
            "log_yvar", mu=np.var(flux_ - sig.medfilt(flux_, 13)), sd=5.0
        )

        # now set up the GP
        gp = GaussianProcess(
            kernel, t=time_, diag=T.exp(log_yvar) * T.ones(len(time_)), mean=mean_
        )
        gp.marginal("gp", observed=flux_)

        # track mean predictions
        full_mean_pred = pm.Deterministic(
            "full_mean_pred", mean_fxn(lc.time, seg, flux0, ramp_amp, log_tau)
        )

    # optimize the GP hyperparameters
    with trend_model:
        trend_map = trend_model.test_point
        trend_map = pmx.optimize(start=trend_map, vars=[flux0], progress=verbose)
        trend_map = pmx.optimize(
            start=trend_map, vars=[flux0, log_yvar], progress=verbose
        )

        for i in range(1 + correct_ramp):
            if kterm == "RotationTerm":
                trend_map = pmx.optimize(
                    start=trend_map,
                    vars=[log_yvar, flux0, sigma, P, log_Q0, log_dQ, mix],
                    progress=verbose,
                )
            if kterm == "SHOTerm":
                trend_map = pmx.optimize(
                    start=trend_map,
                    vars=[log_yvar, flux0, sigma, P, log_Q0],
                    progress=verbose,
                )
            if correct_ramp:
                trend_map = pmx.optimize(
                    start=trend_map,
                    vars=[log_yvar, flux0, ramp_amp, log_tau],
                    progress=verbose,
                )

        trend_map = pmx.optimize(start=trend_map, progress=verbose)

    # reconstruct the GP to interpolate over masked transits
    if kterm == "RotationTerm":
        kernel = GPterms.RotationTerm(
            sigma=trend_map["sigma"],
            period=trend_map["P"],
            Q0=T.exp(trend_map["log_Q0"]),
            dQ=T.exp(trend_map["log_dQ"]),
            f=trend_map["mix"],
        )

    elif kterm == "SHOTerm":
        kernel = GPterms.SHOTerm(
            sigma=trend_map["sigma"],
            w0=2 * pi / trend_map["P"],
            Q=0.5 + T.exp(trend_map["log_Q0"]),
        )

    gp = GaussianProcess(kernel, mean=0.0)
    gp.compute(time_, diag=T.exp(trend_map["log_yvar"]) * T.ones(len(time_)))

    full_trend = (
        gp.predict(flux_ - trend_map["mean_"], lc.time).eval()
        + trend_map["full_mean_pred"]
    )

    lc.flux /= full_trend
    lc.error /= full_trend

    if return_trend:
        return lc, full_trend

    return lc


def stitch(litecurves):
    """
    Combine a list of LiteCurves in a single LiteCurve

    Parameters
    ----------
        litecurves : list of alderaan.LiteCurve() objects

    Returns
    -------
        lc : a single combined alderaan.LiteCurve() object
    """
    combo = deepcopy(litecurves[0])

    for i in range(1, len(litecurves)):
        for k in combo.__dict__.keys():
            if type(combo.__dict__[k]) is np.ndarray:
                combo.__dict__[k] = np.hstack(
                    [combo.__dict__[k], litecurves[i].__dict__[k]]
                )

    return combo


def filter_ringing(lc, break_tolerance, fring, bw):
    """
    Filter out known long cadence instrumental ringing modes (see Gilliland+ 2010)
    Applies a notch filter (narrow bandstop filter) at a set of user specified frequencies

    The function does NOT change the lc.flux attribute directly, but rather returns a new flux array

    Parameters
    ----------
        lc : alderaan.LiteCurve
            Litecurve must have time, flux, and cadno attributes
        break_tolerance : int
            number of cadences considered a large gap in time
        fring : array-like
            ringing frequencies in same units as lc.time (i.e. if time is in days, fring is in days^-1)
        bw : float
            bandwidth of stopband (same units as fring)

    Returns
    -------
        flux_filtered : ndarray
            flux with ringing modes filtered out
    """
    # make lists to hold outputs
    flux_filtered = []

    # identify gaps
    gap_locs = identify_gaps(lc, break_tolerance, jump_tolerance=5.0)

    # break the data into contiguous segments and detrend
    for i, gloc in enumerate(gap_locs[:-1]):
        # grab segments of time, flux, cadno, masks
        t = lc.time[gap_locs[i] : gap_locs[i + 1]]
        f = lc.flux[gap_locs[i] : gap_locs[i + 1]]
        c = lc.cadno[gap_locs[i] : gap_locs[i + 1]]

        # fill small gaps with white noise
        npts = c[-1] - c[0] + 1
        dt = np.min(t[1:] - t[:-1])

        t_interp = np.linspace(t.min(), t.max() + dt * 3 / 2, npts)
        f_interp = np.ones_like(t_interp)
        c_interp = np.arange(c.min(), c.max() + 1)

        data_exists = np.isin(c_interp, c)

        f_interp[data_exists] = f
        f_interp[~data_exists] = np.random.normal(
            loc=np.median(f), scale=np.std(f), size=np.sum(~data_exists)
        )

        # now apply the filter
        f_fwd_back = np.copy(f_interp)
        f_back_fwd = np.copy(f_interp)
        f_ramp = np.linspace(0, 1, len(f_interp))

        for j, f0 in enumerate(fring):
            b, a = sig.iirnotch(f0, Q=2 * f0 / bw, fs=1 / dt)
            f_fwd_back = sig.filtfilt(
                b, a, f_fwd_back, padlen=np.min([120, len(f_fwd_back) - 2])
            )
            f_back_fwd = sig.filtfilt(
                b, a, f_back_fwd[::-1], padlen=np.min([120, len(f_back_fwd / 2) - 1])
            )[::-1]

            f_filt = f_fwd_back * f_ramp + f_fwd_back * f_ramp[::-1]

        flux_filtered.append(f_filt[data_exists])

    return np.hstack(flux_filtered)



__all__ = ["Ephemeris"]


import numpy as np


class Ephemeris:
    """
    A class for storing the emphemeris information of a single planet

    Parameters
    ----------

    Attributes
    ----------


    """

    def __init__(self, inds, tts):
        self.inds = inds
        self.tts = tts

        # calculate least squares period and epoch
        self.period, self.t0 = self._fit_ephem()

        # calculate ttvs
        self.ttvs = self.tts - self.period * self.inds + self.t0

        # calculate full set of transit times
        self.full_transit_times = self.t0 + self.period * np.arange(self.inds.max() + 1)
        self.full_transit_times[self.inds] = self.tts

        # set up histogram for identifying transit offsets
        ftts = self.full_transit_times

        self._bin_edges = np.concatenate(
            [
                [ftts[0] - 0.5 * self.period],
                0.5 * (ftts[1:] + ftts[:-1]),
                [ftts[-1] + 0.5 * self.period],
            ]
        )

        self._bin_values = np.concatenate([[ftts[0]], ftts, [ftts[-1]]])

    def _fit_ephem(self):
        A = np.ones((len(self.inds), 2))
        A[:, 0] = self.inds

        return np.linalg.lstsq(A, self.tts)[0]

    def _get_model_dt(self, t, return_inds=False):
        _inds = np.searchsorted(self._bin_edges, t)
        _vals = self._bin_values[_inds]

        if return_inds:
            return _vals, _inds
        return _vals

    def _warp_times(self, t, return_inds=False):
        warps = self._get_model_dt(t, return_inds=return_inds)

        if return_inds:
            return t - warps[0], warps[1]
        else:
            return t - warps



__all__ = [
    "boxcar_smooth",
    "FFT_estimator",
    "LS_estimator",
    "bin_data",
    "autocorr_length",
    "weighted_percentile",
    "lorentzian",
    "heavyside",
    "draw_random_samples",
    "get_bw",
    "generate_synthetic_samples",
]


import warnings

import aesara_theano_fallback.tensor as T
import astropy.stats
from astropy.timeseries import LombScargle
import numpy as np
from scipy import fftpack
import scipy.optimize as op
import scipy.signal as sig
from scipy import stats
from scipy.interpolate import interp1d
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
from sklearn.covariance import EmpiricalCovariance

from .constants import *


def boxcar_smooth(x, winsize, passes=1):
    """
    Smooth a data array with a sliding boxcar filter

    Parameters
    ----------
    x : ndarray
        data to be smoothed
    winsize : int
        size of boxcar window
    passes : int
        number of passes (default=1)

    Returns
    -------
    xsmooth : ndarray
        smoothed data array,same size as input 'x'
    """
    win = sig.boxcar(winsize) / winsize
    xsmooth = np.pad(x, (winsize, winsize), mode="reflect")

    for i in range(passes):
        xsmooth = sig.convolve(xsmooth, win, mode="same")

    xsmooth = xsmooth[winsize:-winsize]

    return xsmooth


def FFT_estimator(
    x,
    y,
    fmin=None,
    fmax=None,
    crit_fap=0.003,
    nboot=1000,
    return_levels=False,
    max_peaks=2,
):
    """
    Identify significant frequencies in a (uniformly sampled) data series

    Parameters
    ----------
    x : array-like
        1D array of x time domain data values; should be monotonically increasing
    y : array-like
        1D array of corresponding y time domain data values, len(x)
    fmin : float (optional)
        minimum frequency to check; if not provided this will be set to 1/baseline
    fmax : float (optional)
        maximum frequency to check; if not provided this will be set to the Nyquist frequency
    crit_fap : float
        critical false alarm probability for significant signals (default=0.003)
    nboot : int
        number of bootstrap samples for calculating false alarm probabilities (default=1000)
    return_levels : bool
        if True, return the FAP for each frequency in the grid

    Returns
    -------
    xf : ndarray
        1D array of frequency values
    yf : ndarray
        1D array of response values, len(xf); |yf|^2 = power
    freqs : ndarray
        array of significant frequencies
    faps : ndarray
        array of corresponding false alarm probabilities
    levels : ndarray
        array of FAPs for each grid frequency
    """
    # set baseline and Nyquist frequencies
    fbas = 1 / (x.max() - x.min())
    fnyq = 1 / (2 * (x[1] - x[0]))

    if fmin is None:
        fmin = 1.0 * fbas
    if fmax is None:
        fmax = 1.0 * fnyq

    # generate FFT (convolve w/ hann window to reduce spectral leakage)
    w = sig.hann(len(x))

    xf = np.linspace(0, fnyq, len(x) // 2)
    yf = np.abs(fftpack.fft(y * w)[: len(x) // 2])

    keep = (xf >= fmin) * (xf <= fmax)

    xf = xf[keep]
    yf = yf[keep]

    # calculate false alarm probabilities w/ bootstrap test
    yf_max = np.zeros(nboot)

    for i in range(nboot):
        y_shuff = np.random.choice(y, size=len(y), replace=False)
        yf_shuff = np.abs(fftpack.fft(y_shuff * w)[: len(x) // 2])
        yf_max[i] = yf_shuff[keep].max()

    yf_fap = np.zeros(len(xf))

    for i in range(len(xf)):
        yf_fap[i] = 1 - np.sum(yf[i] > yf_max) / nboot

    levels = np.array(
        [
            np.percentile(yf_max, 90),
            np.percentile(yf_max, 99),
            np.percentile(yf_max, 99.9),
        ]
    )

    # now search for significant frequencies
    m = np.zeros(len(xf), dtype="bool")
    freqs = []
    faps = []

    if len(xf) > 3:
        loop = True
        while loop:
            peakfreq = xf[~m][np.argmax(yf[~m])]
            peakfap = yf_fap[~m][np.argmax(yf[~m])]

            if peakfreq == xf[~m].min():
                m[xf <= xf[~m].min()] = 1

            elif peakfap < crit_fap:
                fxn_ = lambda theta, x, y: y - lorentzian(theta, x)

                theta_in = np.array([peakfreq, fbas, yf.max(), np.median(yf)])
                theta_out, success = op.leastsq(fxn_, theta_in, args=(xf, yf))

                width = np.max(5 * [theta_out[1], 3 * (xf[1] - xf[0])])
                m += np.abs(xf - theta_out[0]) / width < 1

                freqs.append(theta_out[0])
                faps.append(peakfap)

            else:
                loop = False

            if len(freqs) >= max_peaks:
                loop = False

            if np.sum(m) / len(m) > 0.5:
                loop = False

    freqs = np.asarray(freqs)
    faps = np.asarray(faps)

    if return_levels:
        return xf, yf, freqs, faps, levels

    return xf, yf, freqs, faps


def LS_estimator(x, y, fsamp=None, fap=0.1, return_levels=False, max_peaks=2):
    """
    Generate a Lomb-Scargle periodogram and identify significant frequencies
      * assumes that data are nearly evenly sampled
      * optimized for finding marginal periodic TTV signals in OMC data
      * may not perform well for other applications

    Parameters
    ----------
    x : array-like
        1D array of x data values; should be monotonically increasing
    y : array-like
        1D array of corresponding y data values, len(x)
    fsamp: float
        nominal sampling frequency; if not provided it will be calculated from the data
    fap : float
        false alarm probability threshold to consider a frequency significant (default=0.1)

    Returns
    -------
    xf : ndarray
        1D array of frequencies
    yf : ndarray
        1D array of corresponding response
    freqs : list
        signficant frequencies
    faps : list
        corresponding false alarm probabilities
    """
    # get sampling frequency
    if fsamp is None:
        fsamp = 1 / np.min(x[1:] - x[:-1])

    # Hann window to reduce ringing
    hann = sig.windows.hann(len(x))
    hann /= np.sum(hann)

    # identify any egregious outliers
    out = np.abs(y - np.median(y)) / astropy.stats.mad_std(y) > 5.0

    xt = x[~out]
    yt = y[~out]

    freqs = []
    faps = []

    loop = True
    while loop:
        lombscargle = LombScargle(xt, yt * hann[~out])
        xf, yf = lombscargle.autopower(
            minimum_frequency=1.5 / (xt.max() - xt.min()),
            maximum_frequency=0.25 * fsamp,
            samples_per_peak=10,
        )

        peak_freq = xf[np.argmax(yf)]
        peak_fap = lombscargle.false_alarm_probability(yf.max(), method="bootstrap")

        # output first iteration of LS periodogram
        if len(freqs) == 0:
            xf_out = xf.copy()
            yf_out = yf.copy()
            levels = lombscargle.false_alarm_level([0.1, 0.01, 0.001])

        if peak_fap < fap:
            yt -= lombscargle.model(xt, peak_freq) * len(xt)
            freqs.append(peak_freq)
            faps.append(peak_fap)

        else:
            loop = False

        if len(freqs) >= max_peaks:
            loop = False

    if return_levels:
        return xf_out, yf_out, freqs, faps, levels

    else:
        return xf_out, yf_out, freqs, faps


def bin_data(time, data, binsize, bin_centers=None):
    """
    Parameters
    ----------
    time : ndarray
        vector of time values
    data : ndarray
        corresponding vector of data values to be binned
    binsize : float
        bin size for output data, in same units as time

    Returns
    -------
    bin_centers : ndarray
        center of each data (i.e. binned time)
    binned_data : ndarray
        data binned to selcted binsize
    """
    if bin_centers is None:
        bin_centers = np.hstack(
            [
                np.arange(time.mean(), time.min() - binsize / 2, -binsize)[::-1],
                np.arange(time.mean(), time.max() + binsize / 2, binsize)[1:],
            ]
        )

    binned_data = np.zeros(len(bin_centers))
    for i, t0 in enumerate(bin_centers):
        binned_data[i] = np.mean(data[np.abs(time - t0) < binsize / 2])

    return bin_centers, binned_data


def autocorr_length(x):
    """
    Determine the autocorrelation length of a 1D data vector

    Parameters
    ----------
    x : array-like
        1D data vector

    Returns
    -------
    tau : float
        estimated autocorrelation length
    """
    # subtract off mean
    y = x - np.mean(x)

    # generate empirical ACF
    acf = np.correlate(y, y, mode="full")
    acf = acf[len(acf) // 2 :]
    acf /= acf[0]

    # automatic windowing following Sokal 1989
    taus = 2 * np.cumsum(acf) - 1

    m = np.arange(len(taus)) < 5.0 * taus
    if np.any(m):
        win = np.argmin(m)
    else:
        win = len(taus) - 1

    return np.max([taus[win], 1.0])


def weighted_percentile(a, q, w=None):
    """
    Compute the q-th percentile of data array a
    Similar to np.percentile, but allows for weights (but not axis-wise computation)
    """
    a = np.array(a)
    q = np.array(q)

    if w is None:
        return np.percentile(a, q)

    else:
        w = np.array(w)

        assert np.all(q >= 0) and np.all(q <= 100), "quantiles should be in [0, 100]"

        order = np.argsort(a)
        a = a[order]
        w = w[order]

        weighted_quantiles = np.cumsum(w) - 0.5 * w
        weighted_quantiles -= weighted_quantiles[0]
        weighted_quantiles /= weighted_quantiles[-1]

        return np.interp(q / 100.0, weighted_quantiles, a)


def lorentzian(theta, x):
    """
    Model a Lorentzian (Cauchy) function

    Parameters
    ----------
    theta : array-like
        parameters for Lorentzian = [loc, scale, height, baseline]
    x : array-like
        1D array of x data values

    Returns
    -------
    pdf : array-like
        Lorentzian probability density function
    """
    loc, scale, height, baseline = theta

    return height * stats.cauchy(loc=loc, scale=scale).pdf(x) + baseline


def heavyside(x, x0=0.0, k=1000.0):
    """
    Approximates the Heavyside step function with a smooth distribution
    Implemented using theano tensors for easily setting bounds on PyMC3 deterministic variables

    H0 = 1/(1+T.exp(-2*k*(x-x0)))

    Parameters
    ----------
    x : array-like
        input values at which to compute function
    x0 : float
        location of step
    k : float (optional)
        exponent factor in approximation (default=1000.)
    """
    return 1 / (1 + T.exp(-2 * k * (x - x0)))


def draw_random_samples(pdf, domain, N, *args, **kwargs):
    """
    Draw random samples from a given pdf using inverse transform sampling

    Paramters
    --------
    pdf : univarite function
        outputs pdf as a function of input value
    domain : tuple
        domain over which pdf is defined
    N : int
        number of random samples to draw

    Returns
    -------
    samples : ndarray
        vector of random samples drawn
    """
    x = np.linspace(domain[0], domain[1], int(1e5))
    y = pdf(x, *args, **kwargs)
    cdf_y = np.cumsum(y)
    cdf_y = cdf_y / cdf_y.max()
    inverse_cdf = interp1d(cdf_y, x)

    return inverse_cdf(np.random.uniform(cdf_y.min(), cdf_y.max(), N))


def get_bw(samples, weights=None, max_draws=1000):
    """
    Use cross-validation to determine KDE bandwidth for a set of 1D data samples
    Iteratively performs first coarse then fine estimation

    Parameters
    ----------
    samples : array-like
        1D data samples
    weights : array-like (optional)
        weights corresponding to samples
    max_draws : int (default=1000)
        maximum number of samples to use during estimation

    Returns
    -------
    bw : float
        estimated bandwidth
    """
    N = len(samples)
    x = np.random.choice(samples, p=weights, size=np.min([max_draws, N]), replace=True)

    coarse_mesh = np.linspace((x.max() - x.min()) / N, np.std(x), int(np.sqrt(N)))
    grid = GridSearchCV(KernelDensity(), {"bandwidth": coarse_mesh}, cv=5)
    grid.fit(x[:, None])

    fine_mesh = grid.best_params_["bandwidth"] + np.linspace(-1, 1, int(np.sqrt(N))) * (
        coarse_mesh[1] - coarse_mesh[0]
    )
    grid = GridSearchCV(KernelDensity(), {"bandwidth": fine_mesh}, cv=5)
    grid.fit(x[:, None])

    return grid.best_params_["bandwidth"]


def generate_synthetic_samples(samples, bandwidths, n_up, weights=None):
    """
    Use PDF Over-Sampling (PDFOS - Gao+ 2014) to generate synthetic samples

    Parameters
    ----------
    samples : ndarray, (N x M)
        array of data samples arranged N_samples, M_parameters
    bandwitdhs : array-like, (M)
        pre-estimated KDE bandwidths for each of M parameters
    n_up : int
        number of upsampled synthetic data points to generate
    weights : ndarray, (N x M)
        array of weights corresponding to samples

    Returns
    -------
    new_samples : ndarray
        array containing synthetic samples
    """
    n_samp, n_dim = samples.shape
    index = np.arange(0, n_samp, dtype="int")

    # we'll generate a few more samples than needed in anticipation of rejecting a few
    n_up101 = int(1.01 * n_up)

    # naive resampling (used only to estimate covariance matrix)
    naive_resamples = samples[np.random.choice(index, p=weights, size=3 * n_samp)]

    # compute empirical covariance matrix and lower Cholesky decomposition
    emp_cov = EmpiricalCovariance().fit(naive_resamples).covariance_
    L = np.linalg.cholesky(emp_cov)

    # scale each parameter by precomputed bandwidths so they have similar variance
    samples_scaled = (samples - np.mean(samples, axis=0)) / bandwidths

    # calculate synthetic samples following PDFOS (Gao+ 2014)
    random_index = np.random.choice(index, p=weights, size=n_up101)
    random_samples = samples_scaled[random_index]
    random_jitter = np.random.normal(0, 1, n_up101 * n_dim).reshape(n_up101, n_dim)
    new_samples = random_samples + np.dot(L.T, random_jitter.T).T

    # rescale each parameter to invert previous scaling
    new_samples = new_samples * bandwidths + np.mean(samples, axis=0)

    # reject any synthetic samples pushed out of bounds of original samples
    bad = np.zeros(n_up101, dtype="bool")

    for i in range(n_dim):
        bad += new_samples[:, i] < samples[:, i].min()
        bad += new_samples[:, i] > samples[:, i].max()

    if np.sum(bad) / len(bad) > 0.01:
        warnings.warn(
            "More than 1% of PDFOS generated samples were beyond min/max values of original samples"
        )

    new_samples = new_samples[~bad]

    # only return n_up samples
    if new_samples.shape[0] >= n_up:
        return new_samples[:n_up]

    else:
        # use naive resampling to replace rejected samples
        replacement_samples = samples[
            np.random.choice(index, p=weights, size=n_up - new_samples.shape[0])
        ]

        return np.vstack([new_samples, replacement_samples])



import os
import sys
sys.path.insert(0, os.path.abspath('./../alderaan/'))

import alderaan

# project info
project = 'alderaan'
copyright = '2025, Gregory J. Gilbert'
authors = 'Gregory J. Gilbert & Erik A. Petigura'
release = alderaan.__version__
version = alderaan.__version__

# general configuration
root_doc = 'index'
templates_path = ['_templates']
html_static_path = ['_static']

html_theme = 'sphinx_rtd_theme'
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



#!/usr/bin/env python
# coding: utf-8

# # Analyze Autocorrelated Noise


import os
import sys
import json
import shutil
import warnings
import argparse
from datetime import datetime
from timeit import default_timer as timer

from aesara_theano_fallback import aesara as theano
from astropy.stats import mad_std
from astropy.units import UnitsWarning
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.polynomial.polynomial as poly
import pymc3_ext as pmx
from scipy import stats


# #### Flush buffer and silence extraneous warnings


# flush buffer to avoid mixed outputs from progressbar
sys.stdout.flush()

# turn off FutureWarnings
warnings.filterwarnings("ignore", category=FutureWarning)

# supress UnitsWarnings (this code doesn't use astropy units)
warnings.filterwarnings(action="ignore", category=UnitsWarning, module="astropy")


# #### Initialize timer


print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Detrending and TTV Estimation")
print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
print("+" * shutil.get_terminal_size().columns)
print("")

# start program timer
global_start_time = timer()


# #### Parse inputs


try:
    parser = argparse.ArgumentParser(
        description="Inputs for ALDERAAN transit fiting pipeline"
    )
    parser.add_argument(
        "--mission",
        default=None,
        type=str,
        required=True,
        help="Mission name; can be 'Kepler' or 'Simulated'",
    )
    parser.add_argument(
        "--target",
        default=None,
        type=str,
        required=True,
        help="Target name; format should be K00000 or S00000",
    )
    parser.add_argument(
        "--run_id", default=None, type=str, required=True, help="run identifier"
    )
    parser.add_argument(
        "--project_dir",
        default=None,
        type=str,
        required=True,
        help="Project directory for accessing lightcurve data and saving outputs",
    )
    parser.add_argument(
        "--data_dir",
        default=None,
        type=str,
        required=True,
        help="Data directory for accessing MAST lightcurves",
    )
    parser.add_argument(
        "--catalog",
        default=None,
        type=str,
        required=True,
        help="CSV file containing input planetary parameters",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable verbose logging",
    )
    parser.add_argument(
        "--iplot",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable interactive matplotlib backend; default 'agg'",
    )

    args = parser.parse_args()
    MISSION = args.mission
    TARGET = args.target
    PROJECT_DIR = args.project_dir
    DATA_DIR = args.data_dir
    CATALOG = args.catalog
    RUN_ID = args.run_id
    VERBOSE = args.verbose
    IPLOT = args.iplot

except SystemExit:
    warnings.warn("No arguments were parsed from the command line")


print("")
print(f"   MISSION : {MISSION}")
print(f"   TARGET  : {TARGET}")
print(f"   RUN ID  : {RUN_ID}")
print("")
print(f"   Project directory : {PROJECT_DIR}")
print(f"   Data directory    : {DATA_DIR}")
print(f"   Input catalog     : {CATALOG}")
print("")
print(f"   theano cache : {theano.config.compiledir}")
print("")


# #### Build directory structure


# directories in which to place pipeline outputs for this run
RESULTS_DIR = os.path.join(PROJECT_DIR, "Results", RUN_ID, TARGET)
FIGURE_DIR = os.path.join(PROJECT_DIR, "Figures", RUN_ID, TARGET)

os.makedirs(RESULTS_DIR, exist_ok=True)
os.makedirs(FIGURE_DIR, exist_ok=True)

sys.path.append(PROJECT_DIR)


# #### Import ALDERAAN routines


import alderaan.io as io
import alderaan.detrend as detrend
import alderaan.noise as noise
from alderaan.astro import make_transit_mask
from alderaan.constants import scit, lcit
from alderaan.plotting import plot_omc, plot_acf, plot_synthetic_noise
from alderaan.utils import boxcar_smooth


# #### Set matplotlib backends


if not IPLOT:
    mpl.use("agg")

if np.any(np.array(["agg", "png", "svg", "pdf", "ps"]) == mpl.get_backend()):
    IPLOT = False


# MAIN SCRIPT BEGINS HERE
def main():
    # # ################
    # # ----- DATA I/O -----
    # # ################

    print("\nLoading data...\n")

    # ## Read in planet and star properties

    path = os.path.join(RESULTS_DIR, f"{TARGET}_transit_parameters.csv")
    catalog = io.parse_catalog(path, MISSION, TARGET)

    KOI_ID = catalog.koi_id.to_numpy()[0]
    KIC_ID = catalog.kic_id.to_numpy()[0]

    NPL = catalog.npl.to_numpy()[0]

    PERIODS = catalog.period.to_numpy()
    DEPTHS = catalog.depth.to_numpy() * 1e-6
    DURS = catalog.duration.to_numpy() / 24

    if np.any(np.diff(PERIODS) <= 0):
        raise ValueError("Planets should be ordered by ascending period")

    # ## Read in detrended lightcurves
    # #### These can be generated by running the script "detrend_and_estimate_ttvs.py"

    if os.path.exists(os.path.join(RESULTS_DIR, f"{TARGET}_lc_detrended.fits")):
        lc = io.load_detrended_lightcurve(
            os.path.join(RESULTS_DIR, f"{TARGET}_lc_detrended.fits")
        )
        lc.season = lc.quarter % 4
    else:
        lc = None

    if os.path.exists(os.path.join(RESULTS_DIR, f"{TARGET}_sc_detrended.fits")):
        sc = io.load_detrended_lightcurve(
            os.path.join(RESULTS_DIR, f"{TARGET}_sc_detrended.fits")
        )
        sc.season = sc.quarter % 4
    else:
        sc = None

    # ## Read in quick transit times
    # #### These can be generated by running the script "detrend_and_estimate_ttvs.py"

    epochs = np.zeros(NPL)
    periods = np.zeros(NPL)
    linear_ephemeris = [None] * NPL

    transit_inds = []
    indep_transit_times = []
    regular_transit_times = []

    for n in range(NPL):
        fname_in = os.path.join(RESULTS_DIR, f"{TARGET}_{n:02d}_quick.ttvs")
        data_in = np.genfromtxt(fname_in)

        transit_inds.append(np.array(data_in[:, 0], dtype="int"))
        indep_transit_times.append(np.array(data_in[:, 1], dtype="float"))
        regular_transit_times.append(np.array(data_in[:, 2], dtype="float"))

        pfit = poly.polyfit(transit_inds[n], regular_transit_times[n], 1)

        epochs[n] = pfit[1]
        periods[n] = pfit[0]
        linear_ephemeris[n] = poly.polyval(transit_inds[n], pfit)

    fig, ax = plot_omc(linear_ephemeris, indep_transit_times, regular_transit_times)
    fig.savefig(os.path.join(FIGURE_DIR, TARGET + f"_omc.png"), bbox_inches="tight")
    if IPLOT:
        plt.show()
    else:
        plt.close()

    # # ####################
    # # --- PRELIMINARIES ---
    # # ####################

    print("\nRunning preliminaries...\n")

    # ## Establish time baseline

    time_min = []
    time_max = []

    if sc is not None:
        time_min.append(sc.time.min())
        time_max.append(sc.time.max())

    if lc is not None:
        time_min.append(lc.time.min())
        time_max.append(lc.time.max())

    TIME_START = np.min(time_min)
    TIME_END = np.max(time_max)

    # put epochs in range (TIME_START, TIME_START + PERIOD)
    for n in range(NPL):
        if epochs[n] < TIME_START:
            adj = 1 + (TIME_START - epochs[n]) // periods[n]
            epochs[n] += adj * periods[n]

        if epochs[n] > (TIME_START + periods[n]):
            adj = (epochs[n] - TIME_START) // periods[n]
            epochs[n] -= adj * periods[n]

    # ## Estimate TTV scatter w/ uncertainty buffer

    ttv_scatter = np.zeros(NPL)
    ttv_buffer = np.zeros(NPL)

    for n in range(NPL):
        # estimate TTV scatter
        ttv_scatter[n] = mad_std(indep_transit_times[n] - regular_transit_times[n])

        # based on scatter in independent times, set threshold so not even one outlier is expected
        N = len(transit_inds[n])
        eta = np.max([3.0, stats.norm.interval((N - 1) / N)[1]])

        ttv_buffer[n] = eta * ttv_scatter[n] + lcit

    # ## Make masks of various widths

    if sc is not None:
        sc_wide_mask = np.zeros((NPL, len(sc.time)), dtype="bool")
        sc_narrow_mask = np.zeros((NPL, len(sc.time)), dtype="bool")
        sc_regular_mask = np.zeros((NPL, len(sc.time)), dtype="bool")

        for n in range(NPL):
            tts = regular_transit_times[n]
            wide_size = np.max([3 / 24, 3.5 * DURS[n]]) + 2 * ttv_buffer[n]
            narrow_size = 0.5 * DURS[n] + ttv_buffer[n]
            regular_size = np.max([3 / 24, 1.5 * DURS[n]]) + 2 * ttv_buffer[n]

            sc_wide_mask[n] = make_transit_mask(sc.time, tts, wide_size)
            sc_narrow_mask[n] = make_transit_mask(sc.time, tts, narrow_size)
            sc_regular_mask[n] = make_transit_mask(sc.time, tts, regular_size)

        sc.mask = sc_regular_mask.sum(0) > 0

    else:
        sc_wide_mask = None
        sc_narrow_mask = None
        sc_regular_mask = None

    if lc is not None:
        lc_wide_mask = np.zeros((NPL, len(lc.time)), dtype="bool")
        lc_narrow_mask = np.zeros((NPL, len(lc.time)), dtype="bool")
        lc_regular_mask = np.zeros((NPL, len(lc.time)), dtype="bool")

        for n in range(NPL):
            tts = regular_transit_times[n]
            wide_size = np.max([3 / 24, 5.5 * DURS[n]]) + 2 * ttv_buffer[n]
            narrow_size = 0.5 * DURS[n] + ttv_buffer[n]
            regular_size = np.max([3 / 24, 1.5 * DURS[n]]) + 2 * ttv_buffer[n]

            lc_wide_mask[n] = make_transit_mask(lc.time, tts, wide_size)
            lc_narrow_mask[n] = make_transit_mask(lc.time, tts, narrow_size)
            lc_regular_mask[n] = make_transit_mask(lc.time, tts, regular_size)

        lc.mask = lc_regular_mask.sum(0) > 0

    else:
        lc_wide_mask = None
        lc_narrow_mask = None

    # ## Determine what data type each season has

    season_dtype = []

    if sc is not None:
        sc_seasons = np.unique(sc.quarter % 4)
    else:
        sc_seasons = np.array([])

    if lc is not None:
        lc_seasons = np.unique(lc.quarter % 4)
    else:
        lc_seasons = np.array([])

    for z in range(4):
        if np.isin(z, sc_seasons):
            season_dtype.append("short")
        elif np.isin(z, lc_seasons):
            season_dtype.append("long")
        else:
            season_dtype.append("none")

    # # ####################################
    # # ----- AUTOCORRELATION ANALYSIS -----
    # # ####################################

    # ## Generate empirical ACF and filter high-frequency ringing

    print("\nGenerating empirical autocorrelation function...\n")
    print("Season data types:", season_dtype, "\n")

    # set cutoff between low and high frequency signals
    fcut = 2 / lcit
    fmin = 2 / (5 * (DURS.max() + lcit))

    # short cadency Nyquist freqency
    fnyq = 1 / (2 * scit)

    # now estimate the ACF
    acf_lag = []
    acf_emp = []
    acf_mod = []
    acf_freqs = []

    for z in range(4):
        if season_dtype[z] == "none":
            acf_lag.append(None)
            acf_emp.append(None)
            acf_mod.append(None)
            acf_freqs.append(None)

        else:
            if season_dtype[z] == "short":
                npts = int(
                    np.min([5 * (1 / 24 + DURS.max()), 2 / 3 * periods.min()]) / scit
                )
                use = sc.season == z
                m_ = sc.mask[use]

                if np.sum(use) > 0:
                    t_ = sc.time[use][~m_]
                    f_ = sc.flux[use][~m_]
                    c_ = sc.cadno[use][~m_]

            if season_dtype[z] == "long":
                npts = int(
                    np.min([5 * (1 / 24 + DURS.max()), 2 / 3 * periods.min()]) / lcit
                )
                use = lc.season == z
                m_ = lc.mask[use]

                if np.sum(use) > 0:
                    t_ = lc.time[use][~m_]
                    f_ = lc.flux[use][~m_]
                    c_ = lc.cadno[use][~m_]

            if np.sum(use) == 0:
                acf_lag.append(None)
                acf_emp.append(None)
                acf_mod.append(None)
                acf_freqs.append(None)

            else:
                # generate the empirical acf (if generate_acf fails, use very low amplitude white noise)
                try:
                    xcor, acor = noise.generate_acf(t_, f_, c_, npts)
                except Exception:
                    try:
                        npts = int(2 / 3 * npts)
                        xcor, acor = noise.generate_acf(t_, f_, c_, npts)
                    except Exception:
                        xcor = 1 + np.arange(npts, dtype="float")
                        acor = (
                            np.random.normal(size=len(xcor))
                            * np.std(f_)
                            * np.finfo(float).eps
                        )

                if season_dtype[z] == "long":
                    xcor = xcor * lcit
                    method = "smooth"
                    window_length = 3

                if season_dtype[z] == "short":
                    xcor = xcor * scit
                    method = "savgol"
                    window_length = None

                # model the acf
                acor_emp, acor_mod, xf, yf, freqs = noise.model_acf(
                    xcor,
                    acor,
                    fcut,
                    fmin=fmin,
                    method=method,
                    window_length=window_length,
                )

                # make some plots
                fig = plot_acf(xcor, acor_emp, acor_mod, xf, yf, freqs, TARGET, z)
                fig.savefig(
                    os.path.join(FIGURE_DIR, TARGET + f"_acf_season_{z}.png"),
                    bbox_inches="tight",
                )
                if IPLOT:
                    plt.show()
                else:
                    plt.close()

                # filter out high-frequency components in short cadence data
                if season_dtype[z] == "short":
                    fring = list(freqs[(freqs > fcut) * (freqs < fnyq)])
                    bw = 1 / (lcit - scit) - 1 / (lcit + scit)

                    if len(fring) > 0:
                        # apply the notch filter
                        flux_filtered = detrend.filter_ringing(sc, 15, fring, bw)

                        # search for addtional ringing frequencies
                        try:
                            xcor, acor = noise.generate_acf(
                                t_, flux_filtered[use][~m_], c_, npts
                            )
                            xcor = xcor * scit
                        except Exception:
                            pass

                        new_freqs = noise.model_acf(
                            xcor, acor, fcut, fmin=fmin, method="savgol"
                        )[4]
                        new_fring = new_freqs[(new_freqs > fcut) * (new_freqs < fnyq)]

                        for nf in new_fring:
                            if np.sum(np.abs(fring - nf) < bw) == 0:
                                fring.append(nf)

                        # re-apply the notch filter with the new list of ringing frequencies
                        flux_filtered = detrend.filter_ringing(sc, 15, fring, bw)

                        # update the LiteCurve
                        sc.flux[use] = flux_filtered[use]
                        f_ = sc.flux[use][~m_]

                    # re-run the ACF modeling on the filtered lightcurve
                    try:
                        xcor, acor = noise.generate_acf(t_, f_, c_, npts)
                        xcor = xcor * scit
                    except Exception:
                        pass

                    acor_emp, acor_mod, xf, yf, freqs = noise.model_acf(
                        xcor, acor, fcut, fmin=fmin, method="savgol"
                    )

                # add to list
                acf_lag.append(xcor)
                acf_emp.append(acor_emp)
                acf_mod.append(acor_mod)
                acf_freqs.append(freqs)

    # ## Save filtered lightcurves

    print("\nSaving detrended lightcurves...\n")

    if lc is not None:
        lc.to_fits(
            TARGET,
            os.path.join(RESULTS_DIR, f"{TARGET}_lc_filtered.fits"),
            cadence="LONG",
        )
    else:
        print("No long cadence data")

    if sc is not None:
        sc.to_fits(
            TARGET,
            os.path.join(RESULTS_DIR, f"{TARGET}_sc_filtered.fits"),
            cadence="SHORT",
        )
    else:
        print("No short cadence data")

    # ## Generate synthetic noise

    print("\nGenerating synthetic noise...\n")

    synth_time = []
    synth_red = []
    synth_white = []

    for z in range(4):
        print(f"SEASON {z} - {season_dtype[z]}")

        if season_dtype[z] == "none":
            synth_time.append(None)
            synth_red.append(None)
            synth_white.append(None)

        else:
            if season_dtype[z] == "short":
                npts = np.min([int(2 * DURS.max() / scit), len(acf_emp[z])])
                use = sc.season == z
                m_ = sc.mask[use]

                if np.sum(use) > 0:
                    t_ = sc.time[use][~m_]
                    f_ = sc.flux[use][~m_]

            if season_dtype[z] == "long":
                npts = np.min([int(5 * DURS.max() / lcit), len(acf_emp[z])])
                use = lc.season == z
                m_ = lc.mask[use]

                if np.sum(use) > 0:
                    t_ = lc.time[use][~m_]
                    f_ = lc.flux[use][~m_]

            if np.sum(use) == 0:
                synth_time.append(None)
                synth_red.append(None)
                synth_white.append(None)

            else:
                if season_dtype[z] == "long":
                    vector_length = 5 * len(acf_lag[z])
                if season_dtype[z] == "short":
                    vector_length = 2 * len(acf_lag[z])

                # pull and split high/low frequencies
                freqs = np.copy(acf_freqs[z])

                low_freqs = freqs[freqs <= fcut]
                high_freqs = freqs[freqs > fcut]

                # generate some synthetic correlated noise
                clipped_acf = (acf_mod[z][:npts]) * np.linspace(1, 0, npts)

                x, red_noise, white_noise = noise.generate_synthetic_noise(
                    acf_lag[z][:npts], clipped_acf, vector_length, np.std(f_)
                )

                # hacky fix for zero red noise
                if np.var(red_noise) == 0:
                    red_noise = 1e-6 * boxcar_smooth(white_noise, 5)

                # add to list
                synth_time.append(x)
                synth_red.append(red_noise)
                synth_white.append(white_noise)

                # plot the synethic noise
                fig = plot_synthetic_noise(x, white_noise, red_noise, TARGET, z, DEPTHS)
                # plt.savefig(os.path.join(FIGURE_DIR, TARGET + f'_synthetic_noise_season_{z}.png'), bbox_inches='tight')
                if IPLOT:
                    plt.show()
                else:
                    plt.close()

    # ## Fit a GP to the synthetic noise

    print("\nFitting a GP to synthetic noise...\n")

    gp_priors = []

    for z in range(4):
        print(f"SEASON {z}")

        if season_dtype[z] == "none":
            gp_priors.append(None)

        else:
            srz = synth_red[z]

            # pull and split high/low frequencies
            freqs = np.copy(acf_freqs[z])

            if freqs is not None:
                low_freqs = freqs[freqs <= fcut]
                high_freqs = freqs[freqs > fcut]

                if len(low_freqs) > 0:
                    lf = low_freqs[0]
                else:
                    lf = None

                if len(high_freqs) > 0:
                    warnings.warn("there are remaining high-frequency noise components")

            else:
                lf = None

            # fit a GP model to the synthetic noise
            try:
                gp_model = noise.build_sho_model(
                    synth_time[z],
                    srz + np.random.normal(srz) * np.std(srz) * 0.1,
                    var_method="fit",
                    fmax=2 / lcit,
                    f0=lf,
                )

                with gp_model:
                    gp_map = gp_model.test_point

                    for mv in gp_model.vars:
                        gp_map = pmx.optimize(start=gp_map, vars=[mv], verbose=VERBOSE)

                    gp_map = pmx.optimize(start=gp_map, verbose=VERBOSE)

                    try:
                        gp_trace = pmx.sample(
                            tune=6000,
                            draws=1500,
                            start=gp_map,
                            chains=2,
                            target_accept=0.9,
                            progressbar=VERBOSE,
                        )
                    except Exception:
                        gp_trace = pmx.sample(
                            tune=12000,
                            draws=1500,
                            start=gp_map,
                            chains=2,
                            target_accept=0.95,
                            progressbar=VERBOSE,
                        )

            except Exception:
                gp_model = noise.build_sho_model(
                    synth_time[z],
                    srz + np.random.normal(srz) * np.std(srz) * 0.1,
                    var_method="local",
                    fmax=2 / lcit,
                    f0=lf,
                    Q0=1 / np.sqrt(2),
                )

                with gp_model:
                    gp_map = gp_model.test_point

                    for mv in gp_model.vars:
                        gp_map = pmx.optimize(start=gp_map, vars=[mv], verbose=VERBOSE)

                    gp_map = pmx.optimize(start=gp_map, verbose=VERBOSE)

                    try:
                        gp_trace = pmx.sample(
                            tune=12000,
                            draws=1500,
                            start=gp_map,
                            chains=2,
                            target_accept=0.95,
                            progressbar=VERBOSE,
                        )
                    except Exception:
                        gp_trace = gp_map

            # track the priors
            gp_priors.append(noise.make_gp_prior_dict(gp_trace))

    # ## Save GP posteriors for use as priors during simulatenous transit fit

    print("\nSaving GP posteriors...\n")

    for z in range(4):
        if gp_priors[z] is not None:
            for k in gp_priors[z].keys():
                gp_priors[z][k] = list(gp_priors[z][k])

            fname_out = os.path.join(RESULTS_DIR, f"{TARGET}_noise_gp_priors_{z}.txt")

            with open(fname_out, "w") as file:
                json.dump(gp_priors[z], file)

    # ## Exit program

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print(
        f"Analysis of autocorrelated noise complete {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}"
    )
    print(f"Total runtime = {(timer()-global_start_time)/60:.1f} min")
    print("+" * shutil.get_terminal_size().columns)


if __name__ == "__main__":
    main()



import os
import sys

import numpy as np
import warnings
from alderaan.ephemeris import Ephemeris

warnings.simplefilter('always', UserWarning)


period = 11.3
epoch = 6.4

index = np.array([1,2,6,7,9])
ttime = epoch + period*index + 0.1*np.random.normal(size=len(index))
error = 0.1*np.ones_like(ttime)

# TTV init
ephemeris = Ephemeris(index=index, ttime=ttime, error=error)

print(ephemeris.period, ephemeris.epoch)
print(len(ephemeris.index), len(ephemeris.ttime))

# linear init
ephemeris = Ephemeris(period=period, epoch=epoch, t_min=0., t_max=1400.)

print(ephemeris.period, ephemeris.epoch)
print(len(ephemeris.index), len(ephemeris.ttime))

print("\npassing")


import os
import sys

import argparse
from configparser import ConfigParser
from pathlib import Path
import warnings
from alderaan.utils.io import resolve_config_path

warnings.simplefilter('always', UserWarning)

base_path = Path(__file__).resolve().parents[1]

# read inputs from config
parser = argparse.ArgumentParser()
parser.add_argument('-m', '--mission', required=True, type=str, help='Observing mission')
parser.add_argument('-t', '--target', required=True, type=str, help='Target ID for star')
parser.add_argument('-c', '--config', required=True, type=str, help='Path to config file')
args = parser.parse_args()

config = ConfigParser()
config.read(args.config)

alderaan_base_path = Path(__file__).resolve().parents[2]
for key, value in config["PATHS"].items():
    config['PATHS'][key] = resolve_config_path(config['PATHS'][key], alderaan_base_path)

mission = args.mission
target = args.target
run_id = config['RUN']['run_id']

data_dir =  config['PATHS']['data_dir']
outputs_dir = config['PATHS']['outputs_dir']
catalog_dir = config['PATHS']['catalog_dir']

catalog_csv = os.path.join(catalog_dir, str(config['ARGS']['catalog_csv']))

print("")
print(f"   MISSION : {mission}")
print(f"   TARGET  : {target}")
print(f"   RUN ID  : {run_id}")
print("")
print(f"   Data directory    : {data_dir}")
print(f"   Config file       : {args.config}")
print(f"   Input catalog     : {os.path.basename(catalog_csv)}")

print("\npassing")


__all__ = ['TransitModel',
           'ShapeTransitModel',
           'TTimeTransitModel',
          ]

import os
import sys

from batman import _quadratic_ld
from batman import _rsky
from copy import deepcopy
import dynesty
from dynesty.utils import print_fn
import io
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import numpy as np
import numpy.polynomial.polynomial as poly
from scipy.optimize import least_squares
from scipy.special import erfinv
import time

from alderaan.modules.base import BaseAlg
from alderaan.utils.astro import bin_data, estimate_transit_depth
from alderaan.utils.stats import uniform_ppf, loguniform_ppf, norm_ppf


class TransitModel(BaseAlg):
    def __init__(self, litecurve, planets, limbdark):
        super().__init__(litecurve, planets)
        
        self.limbdark = limbdark
        self._init_time_warping()


    def _init_time_warping(self):
        """
        This function initializes warped time arrays
        * the litecurve.time vector is copied for each of the N planets
        * times are then "warped" to account for transit timing variations
        * the transit model assumes linear perturbations to a fixed ephemeris
            
        IMPORTANT!!! The warp functions assume zero-indexing on transit indexes
        """
        # track static ephemerides
        self._static_period = np.array([p.ephemeris._static_period for p in self.planets])
        self._static_epoch = np.array([p.ephemeris._static_epoch for p in self.planets])

        # set warping bins
        self._set_bins()

        # warp lc.time vector for ecah planet; legx is 1st-order Legendre polynomial
        self._warped_time = [None]*self.npl
        self._warped_legx = [None]*self.npl

        for n, p in enumerate(self.planets):
            index_midpoint = p.ephemeris.index[-1] // 2

            _t, _i = self._warp_times(self.litecurve.time, n, return_inds=True)            
            _x = (_i -index_midpoint) / index_midpoint

            self._warped_time[n] = _t.copy()   # len(self.litecurve.ttime)
            self._warped_legx[n] = _x.copy()   # len(self.litecurve.ttime)


    def _set_bins(self):
        self._bin_edges = [None]*self.npl
        self._bin_values = [None]*self.npl

        for n, p, in enumerate(self.planets):
            index_full = np.arange(0, p.ephemeris.index.max()+1, dtype=int)
            ttime_full = self._static_epoch[n] + self._static_period[n] * index_full

            self._bin_edges[n] = np.concatenate(
                [
                    [ttime_full[0] - 0.5 * self._static_period[n]],
                    0.5 * (ttime_full[1:] + ttime_full[:-1]),
                    [ttime_full[-1] + 0.5 * self._static_period[n]],
                ]
            )

            self._bin_values[n] = np.concatenate([[ttime_full[0]], ttime_full, [ttime_full[-1]]])


    def _get_model_dt(self, t, planet_no, return_inds=False):
        _inds = np.searchsorted(self._bin_edges[planet_no], t)
        _vals = self._bin_values[planet_no][_inds]

        if return_inds:
            return _vals, _inds
        return _vals


    def _warp_times(self, t, planet_no, return_inds=False):
        warps = self._get_model_dt(t, planet_no, return_inds=return_inds)

        if return_inds:
            return t - warps[0], warps[1]
        else:
            return t - warps


class ShapeTransitModel(TransitModel):
    def __init__(self, litecurve, planets, limbdark):
        super().__init__(litecurve, planets, limbdark)
        

    @staticmethod
    def model_flux(theta, transitmodel):
        """
        theta : array-like
            num_planets * [C0, C1, rp, b, T14] + [q1, q2]
        transitmodel : TransitModel
            instance of TransitModel, typically self

        Priors on planet parameters [C0, C1, rp, b, T14] are enforced outside this function
        
        Limb darkening is sampled in terms of [q1, q2] (see Kipping 2013)
        Limb darkening priors are enforced in terms of [u1, u2] inside this function
        """
        # shorthand
        tm = transitmodel
        lc = tm.litecurve

        # physical limb darkening (see Kipping 2013)
        q1, q2 = np.array(theta[-2:])
        u1 = 2 * np.sqrt(q1) * q2
        u2 = np.sqrt(q1) * (1 - 2 * q2)

        # calculate log-likelihood
        flux_mod = np.ones_like(lc.flux)

        for obsmode in tm.unique_obsmodes:
            exptime_ioff = tm._exptime_integration_offset_lookup[obsmode]
            supersample = tm._supersample_lookup[obsmode]

            for n, p in enumerate(tm.planets):
                C0, C1, rp, b, T14 = np.array(theta[5 * n : 5 * (n + 1)])

                _t = tm._warped_time[n] + C0 + C1 * tm._warped_legx[n]
                _t_supersample = (exptime_ioff + _t.reshape(_t.size, 1)).flatten()

                nthreads = 1
                ds = _rsky._rsky(
                    _t_supersample,
                    0.0,
                    tm._static_period[n],
                    rp,
                    b,
                    T14,
                    1,
                    nthreads,
                )

                qld_flux = _quadratic_ld._quadratic_ld(
                    ds, np.abs(rp), u1, u2, nthreads
                )

                qld_flux = np.mean(
                    qld_flux.reshape(-1, supersample), axis=1
                )

                flux_mod[lc.obsmode == obsmode] += qld_flux - 1.0
                
        return flux_mod      
                
                
    @staticmethod
    def model_residuals(theta, transitmodel):
        flux_mod = transitmodel.model_flux(theta, transitmodel)
        flux_obs = transitmodel.litecurve.flux
        flux_err = transitmodel.litecurve.error

        return (flux_obs - flux_mod) / flux_err

    
    @staticmethod
    def ln_likelihood(theta, transitmodel):
        # shorthand
        tm = transitmodel

        # likelihood
        lnlike = -0.5 * np.sum(tm.model_residuals(theta, tm)**2)
               
        # enforce prior on limb darkening
        sig_ld_sq = 0.01
        lnlike -= 1.0 / (2 * sig_ld_sq) * (theta[-1] - tm.limbdark[0]) ** 2
        lnlike -= 1.0 / (2 * sig_ld_sq) * (theta[-2] - tm.limbdark[1]) ** 2

        if not np.isfinite(lnlike):
            return -1e300
        
        return lnlike
    

    def sample(self, checkpoint_file=None, checkpoint_every=60, progress_every=10):
        ndim = 5 * self.npl + 2
        sampler = dynesty.DynamicNestedSampler(
            self.ln_likelihood, 
            prior_transform,
            ndim,
            bound='multi',
            sample='rwalk',
            logl_args=(self,),
            ptform_args=(self.durs,),
        )
        sampler.run_nested(
            checkpoint_file=checkpoint_file,
            checkpoint_every=checkpoint_every,
            print_progress=True,
            print_func=throttled_print_fn(progress_every),
        )
        
        return sampler.results
    

    def _theta_initial(self):
        """
        theta : array-like
            num_planets * [C0, C1, rp, b, T14] + [q1, q2]
        """
        theta = []
        for n, p in enumerate(self.planets):
            theta.append(0.0)
            theta.append(0.0)
            theta.append(np.sqrt(p.depth))
            theta.append(0.5)
            theta.append(p.duration)

        theta.append(self.limbdark[0])
        theta.append(self.limbdark[1])

        return np.array(theta)
    

    def _theta_bounds(self):
        """
        theta : array-like
            num_planets * [C0, C1, rp, b, T14] + [q1, q2]
        """
        bounds = []
        for n, p in enumerate(self.planets):
            bounds.append([-np.inf,np.inf])
            bounds.append([-np.inf,np.inf])
            bounds.append([1e-5,0.99])
            bounds.append([0.0,1.0])
            bounds.append([0.01*p.duration,3*p.duration])

        bounds.append([0.,1.])
        bounds.append([0.,1.])

        return np.array(bounds)

    
    def optimize(self, fix_limbdark=True, niter=3):
        theta_initial = self._theta_initial()
        bounds = self._theta_bounds()

        def _fxn(x, x0, fix, self):
            _x = x0.copy()
            _x[~fix] = x
            return self.model_residuals(_x, self)
        
        i_fix = np.arange(len(theta_initial), dtype=int)
        var_names = np.array('C0 C1 r b T14'.split())

        # fix ephemeris, vary [r, b, T14]
        fix_C0_C1 = np.zeros(len(i_fix), dtype=bool)
        fix_C0_C1[i_fix % 5 == 0] = True
        fix_C0_C1[i_fix % 5 == 1] = True

        # fix [r, b, T14], vary ephemeris
        fix_r_b_T14 = np.zeros(len(i_fix), dtype=bool)
        fix_r_b_T14[i_fix % 5 == 2] = True
        fix_r_b_T14[i_fix % 5 == 3] = True
        fix_r_b_T14[i_fix % 5 == 4] = True

        # limb darkening
        fix_C0_C1[-2:] = fix_limbdark
        fix_r_b_T14[-2:] = fix_limbdark

        # do the optimization
        theta_final = theta_initial.copy()

        for fix in [fix_r_b_T14, fix_C0_C1] * niter:
            print(f"optimizing logp for variables: [{', '.join(var_names[~fix[:5]])}]")

            logp_initial = self.ln_likelihood(theta_final, self).copy()

            result = least_squares(
                _fxn,
                theta_final[~fix], 
                method='trf',
                bounds=bounds[~fix,:].T,
                args=[theta_initial, fix, self],
            )

            theta_final[~fix] = result.x.copy()
            logp_final = self.ln_likelihood(theta_final, self).copy()

            print(f"logp: {logp_initial} -> {logp_final}")
            print(f"message: {result['message']}")

        return theta_final
    

    def update_planet_parameters(self, theta):
        for n, p in enumerate(self.planets):
            #self.planets[n].period = self.planets[n].period
            #self.planets[n].epoch = self.planets[n].epoch
            self.planets[n].ror = theta[2 + n * 5]
            self.planets[n].impact = theta[3 + n * 5]
            self.planets[n].duration = theta[4 + n * 5]
            self.planets[n].depth = estimate_transit_depth(p.ror, p.impact)

        return self.planets
    

    def update_limbdark_parameters(self, theta):
        self.limbdark = theta[-2:]

        return self.limbdark
    

class TTimeTransitModel(TransitModel):
    def __init__(self, litecurve, planets, limbdark):
        super().__init__(litecurve, planets, limbdark)


    def _construct_template(self, planet_no, obsmode, transit_window_size):
        """
        planet_no (int) : planet number index

        IMPORTANT!!! Planet extensions must have desired [P, Rp/Rs, b, T14]
        """
        # shorthand
        p = self.planets[planet_no]

        exptime = self._exptime_lookup[obsmode]
        supersample = self._supersample_lookup[obsmode]

        time_template = np.arange(0, transit_window_size/2, exptime/supersample/2, dtype=float)
        time_template = np.hstack([-time_template[:-1][::-1], time_template])
        flux_template = np.zeros_like(time_template)

        nthreads = 1
        ds = _rsky._rsky(
            time_template,
            0.0,
            p.period,
            p.ror,
            p.impact,
            p.duration,
            1,
            nthreads,
        )

        qld_flux = _quadratic_ld._quadratic_ld(
            ds, np.abs(p.ror), self.limbdark[0], self.limbdark[1], nthreads
        )

        flux_template += qld_flux

        return time_template, flux_template
    

    def mazeh13_holczer16_method(
            self, planet_no, rel_window_size=5.0, abs_window_size_buffer=1/24, quicklook_dir=None,
        ):
        """
        Docstring
        """
        # shorthand
        lc = self.litecurve
        p = self.planets[planet_no]

        ttime = np.nan * np.ones(len(p.ephemeris.ttime))
        ttime_err = np.nan * np.ones(len(p.ephemeris.ttime))
        
        overlap = self.identify_overlapping_transits(rtol=1.0, atol=1.0)
       
        assert len(p.ephemeris.ttime) == len(overlap[planet_no]), (
            f"Mismatched sizes for ttime ({len(p.ephemeris.ttime)}) and overlap ({len(overlap[planet_no])})"
        )

        transit_obsmode = self.transit_obsmode[planet_no]

        for obsmode in self.unique_obsmodes:
            print(f"  Determining transit times using {obsmode} data")

            exptime = self._exptime_lookup[obsmode]
            supersample = self._supersample_lookup[obsmode]
            exptime_ioff = self._exptime_integration_offset_lookup[obsmode]

            transit_window_size = rel_window_size * p.duration + abs_window_size_buffer
            time_template, flux_template = self._construct_template(planet_no, obsmode, transit_window_size)

            
            for j, tc in enumerate(p.ephemeris.ttime):
                if (not overlap[planet_no][j]) and (p.ephemeris.quality[j]) and (transit_obsmode[j] == obsmode):
                    #print(f"  Transit {p.ephemeris.index[j]} : BKJD = {tc:.1f}")
                    
                    # STEP 0: pull data near a single transit
                    in_transit = np.abs(self.litecurve.time - tc) < p.duration / 2
                    in_window = np.abs(self.litecurve.time - tc) < transit_window_size / 2
                    
                    if np.sum(in_transit) > 0:
                        _t = lc.time[in_window]
                        _f_obs = lc.flux[in_window]
                        _f_err = lc.error[in_window]

                        _t_supersample = (exptime_ioff + _t.reshape(_t.size, 1)).flatten()

                        # remove any residual out-of-transit trend
                        use = ~in_transit[in_window]
                        try:
                            _f_obs /= poly.polyval(_t, poly.polyfit(_t[use], _f_obs[use], 1))
                        except TypeError:
                            pass

                        # STEP 1: generate tc_offset vs chisq vectors
                        gridstep = exptime / supersample / 1.618 / 3
                        tc_offset = np.arange(0, transit_window_size / 2, gridstep)
                        tc_offset = tc + np.hstack([-tc_offset[:-1][::-1], tc_offset])
                        chisq = np.zeros_like(tc_offset)

                        for i, tc_o in enumerate(tc_offset):
                            _f_mod = np.interp(_t_supersample - tc_o, time_template, flux_template)
                            _f_mod = bin_data(_t_supersample, _f_mod, exptime, bin_centers=_t)[1]

                            chisq[i] = np.sum(((_f_obs - _f_mod) / _f_err)**2)

                        # STEP 2: isolate relevant portions of {tc_offset, chisq} vectors
                        delta_chisq = 2.0

                        loop = True
                        while loop:
                            # grab data near chisq minimum
                            tc_fit = tc_offset[chisq < chisq.min() + delta_chisq]
                            x2_fit = chisq[chisq < chisq.min() + delta_chisq]

                            # eliminate points far from the local minimum
                            faraway = np.abs(tc_fit - np.median(tc_fit)) / np.median(np.diff(tc_fit)) > 1 + 0.5*len(tc_fit)

                            tc_fit = tc_fit[~faraway]
                            x2_fit = x2_fit[~faraway]

                            # check stopping conditions
                            if len(x2_fit) > 7:
                                loop = False
                            if delta_chisq >= 16:
                                loop = False

                            # increment chisq
                            delta_chisq *= np.sqrt(2)

                        # STEP 3: fit a parabola to local chisq minimum
                        if len(tc_fit) > 3:
                            # polynomial fitting
                            quad_coeffs = np.polyfit(tc_fit, x2_fit, 2)
                            quad_model = np.polyval(quad_coeffs, tc_fit)
                            qtc_min = -quad_coeffs[1] / (2 * quad_coeffs[0])
                            qx2_min = np.polyval(quad_coeffs, qtc_min)
                            qtc_err = np.sqrt(1 / quad_coeffs[0])

                            # transit time and scaled error
                            _ttj = np.nanmean([qtc_min, np.mean(tc_fit)])
                            _errj = qtc_err * (1 + np.std(x2_fit - quad_model))

                            # check that the fit is well-conditioned
                            convex_local_min = quad_coeffs[0] > 0
                            within_bounds = (_ttj > tc_fit.min()) and (_ttj < tc_fit.max())

                            #if convex_local_min and within_bounds:
                            ttime[j] = _ttj.copy()
                            ttime_err[j] = _errj.copy()

                        # STEP 4: make quicklook plot
                        if (quicklook_dir is not None) and (len(_f_obs) > 0):
                            target = 'K00148'
                            ttv_dir = os.path.join(quicklook_dir, 'ttvs')
                            os.makedirs(ttv_dir, exist_ok=True)
                            path = os.path.join(ttv_dir, f'{target}_{planet_no}_ttv_{j}.png')

                            _f_mod = np.interp(_t_supersample - _ttj, time_template, flux_template)
                            _f_mod = bin_data(_t_supersample, _f_mod, exptime, bin_centers=_t)[1]
                            
                            fig, ax = plt.subplots(1,2, figsize=(8,3))
                            
                            ax[0].plot(_t, _f_obs, 'ko')
                            ax[0].plot(_t, _f_mod, c=f'C{planet_no}', lw=3)

                            xticks = np.array([tc-transit_window_size/2, tc, tc+transit_window_size/2]).round(2)
                            ax[0].set_xticks(xticks)
                            ax[0].yaxis.set_major_formatter(FormatStrFormatter('%.4f'))
                            ax[0].set_xlabel("Time [BJKD]", fontsize=14)
                            ax[0].set_ylabel("Flux", fontsize=14)

                            display = np.abs(chisq - qx2_min) < 2.5

                            _x = tc_offset[display]
                            _y_obs = (chisq-qx2_min)[display]
                            _y_mod = np.polyval(quad_coeffs, _x) - qx2_min

                            ax[1].plot(_x, _y_obs, 'o', mec='k', mfc='w')
                            ax[1].plot(_x, _y_mod, c=f'C{planet_no}', lw=3)
                            ax[1].axvline(qtc_min, color='k', ls=':')
                            ax[1].axvline(tc_fit[np.argmin(x2_fit)], color='k', ls=':')
                            ax[1].axvline(np.mean(tc_fit), color='k', ls=':')
                            ax[1].axvline(_ttj, color='k', lw=2)

                            xticks = np.array([_ttj - 1.5 * _errj, _ttj, _ttj + 1.5 * _errj])
                            ax[1].set_xticks(xticks, np.round(xticks - _ttj, 4))
                            ax[1].set_ylim(-0.5, 2.5)
                            ax[1].set_xlabel("$\Delta t_c$", fontsize=14)
                            ax[1].set_ylabel("$\Delta \chi^2$", fontsize=14)
                            
                            plt.suptitle(f"{target} - Planet {planet_no}", fontsize=18)
                            plt.tight_layout()
                            fig.savefig(path)
                            plt.close()

        return ttime, ttime_err
    

def prior_transform(uniform_hypercube, fixed_durations):
    """
    Prior transform over physical ALDERAAN basis {C0, C1, r, b, T14}...{q1,q2}

    Distributions are hard-coded to be:
        Normal on ephemeris perturbations {C0,C1}
        Log-uniform on radius ratio r and transit duration T14
        Uniform on impact parameter b
        Uniform on quadratic limb darkening coefficients {q1,q2}

    For motivation behind this prior-parameter choice see:
        Carter+ 2008 (2008ApJ...689..499C)
        Kipping 2013 (2013MNRAS.435.2152K)
        Gilbert, MacDougall, & Petigura 2022 (2022AJ....164...92G)
        MacDougal, Gilbert, & Petigura 2023 (2023AJ....166...61M)

    Assumes that transit durations are known with reasonable confidence
      * sampled T14 will be restricted between (0.01*fixed_duration, 3*fixed_duration)

    Parameters
    ----------
        uniform_hypercube : array-like, length 5 * N + 2
            list of parameters N * [C0, C1, r, b, T14] + [q1, q2]
        fixed_durations : array-like
            list of transit durations for N planets (used for setting prior limits)

    Returns
    -------
        transformed_hypercube : array_like, length 5 * N + 2
            transformed samples
    """
    npl = len(fixed_durations)

    u = np.array(uniform_hypercube)  # U(0,1) priors
    x = np.zeros_like(u)             # physical priors

    # 5 * npl + 2 parameters: [C0, C1, r, b, T14]...[q1, q2]
    for n in range(npl):
        x[5 * n + 0] = norm_ppf(u[0 + n * 5], 0.0, 0.1)
        x[5 * n + 1] = norm_ppf(u[1 + n * 5], 0.0, 0.1)
        x[5 * n + 2] = loguniform_ppf(u[2 + n * 5], 1e-5, 0.99)
        x[5 * n + 3] = uniform_ppf(u[3 + n * 5], 0.0, 1 + x[5 * n + 2])
        x[5 * n + 4] = loguniform_ppf(u[4 + n * 5], 0.01*fixed_durations[n], 3 * fixed_durations[n])

    # limb darkening coefficients (see Kipping 2013)
    x[-2] = uniform_ppf(u[-2], 0, 1)
    x[-1] = uniform_ppf(u[-1], 0, 1)

    return x


def throttled_print_fn(interval=10):
    last = [0]
    start = time.time()

    def callback(results, niter, ncall, *args, **kwargs):
        now = time.time()
        if now - last[0] >= interval:
            runtime = (now - start)/60

            buf = io.StringIO()
            stdout = sys.stdout
            sys.stdout = buf
            print_fn(results, niter, ncall, *args, **kwargs)
            sys.stdout = stdout

            line = buf.getvalue().rstrip()
            print(f"{line} | runtime: {runtime:.1f} min")
            last[0] = now

    return callback


import os
import sys

base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../..'))
if base_path not in sys.path:
    sys.path.insert(0, base_path)

import argparse
from astropy.units import UnitsWarning
from configparser import ConfigParser
from datetime import datetime
import shutil
from src.utils.io import expand_config_path, copy_input_target_catalog
from timeit import default_timer as timer
import warnings

def execute():
    # flush buffer
    sys.stdout.flush()
    sys.stderr.flush()

    # filter warnings
    warnings.simplefilter('always', UserWarning)
    warnings.filterwarnings(
        action='ignore', category=UnitsWarning, module='astropy'
    )

    # start timer
    global_start_time = timer()

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print("ALDERAAN Pipeline")
    print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
    print("+" * shutil.get_terminal_size().columns)
    print("")

    # read inputs from config
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--mission', required=True, type=str, help='Observing mission')
    parser.add_argument('-t', '--target', required=True, type=str, help='Target ID for star')
    parser.add_argument('-c', '--config', required=True, type=str, help='Path to config file')
    args = parser.parse_args()

    config = ConfigParser()
    config.read(os.path.join(base_path, args.config))

    mission = args.mission
    target = args.target
    run_id = config['RUN']['run_id']

    data_dir =  expand_config_path(config['PATHS']['data_dir'])
    outputs_dir = expand_config_path(config['PATHS']['outputs_dir'])
    catalog_dir = expand_config_path(config['PATHS']['catalog_dir'])

    catalog_csv = os.path.join(catalog_dir, str(config['ARGS']['catalog_csv']))

    print("")
    print(f"   MISSION : {mission}")
    print(f"   TARGET  : {target}")
    print(f"   RUN ID  : {run_id}")
    print("")
    print(f"   Base path         : {base_path}")
    print(f"   Data directory    : {data_dir}")
    print(f"   Config file       : {args.config}")
    print(f"   Input catalog     : {os.path.basename(catalog_csv)}")
    print("")
    print(f"   theano cache : {theano.config.compiledir}")
    print("")

    # build directory structure
    os.makedirs(outputs_dir, exist_ok=True)

    results_dir = os.path.join(outputs_dir, 'results', run_id, target)
    os.makedirs(results_dir, exist_ok=True)

    quicklook_dir = os.path.join(outputs_dir, 'quicklook', run_id, target)
    os.makedirs(quicklook_dir, exist_ok=True)

    # copy input catalog into results directory
    catalog_csv_copy = os.path.join(outputs_dir, 'results', run_id, f'{run_id}.csv')
    copy_input_target_catalog(catalog_csv, catalog_csv_copy)


if __name__ == '__execute__':
    execute()


import argparse
import os
import stat


def lookup_epochs(quarter, cadence):
    LONG_QUARTER_PREFIXES = {
        "0": ["2009131105131"],
        "1": ["2009166043257"],
        "2": ["2009259160929"],
        "3": ["2009350155506"],
        "4": ["2010078095331", "2010009091648"],
        "5": ["2010174085026"],
        "6": ["2010265121752"],
        "7": ["2010355172524"],
        "8": ["2011073133259"],
        "9": ["2011177032512"],
        "10": ["2011271113734"],
        "11": ["2012004120508"],
        "12": ["2012088054726"],
        "13": ["2012179063303"],
        "14": ["2012277125453"],
        "15": ["2013011073258"],
        "16": ["2013098041711"],
        "17": ["2013131215648"],
    }

    SHORT_QUARTER_PREFIXES = {
        "0": ["2009131110544"],
        "1": ["2009166044711"],
        "2": ["2009201121230", "2009231120729", "2009259162342"],
        "3": ["2009291181958", "2009322144938", "2009350160919"],
        "4": ["2010009094841", "2010019161129", "2010049094358", "2010078100744"],
        "5": ["2010111051353", "2010140023957", "2010174090439"],
        "6": ["2010203174610", "2010234115140", "2010265121752"],
        "7": ["2010296114515", "2010326094124", "2010355172524"],
        "8": ["2011024051157", "2011053090032", "2011073133259"],
        "9": ["2011116030358", "2011145075126", "2011177032512"],
        "10": ["2011208035123", "2011240104155", "2011271113734"],
        "11": ["2011303113607", "2011334093404", "2012004120508"],
        "12": ["2012032013838", "2012060035710", "2012088054726"],
        "13": ["2012121044856", "2012151031540", "2012179063303"],
        "14": ["2012211050319", "2012242122129", "2012277125453"],
        "15": ["2012310112549", "2012341132017", "2013011073258"],
        "16": ["2013017113907", "2013065031647", "2013098041711"],
        "17": ["2013121191144", "2013131215648"],
    }
    if cadence == "long":
        if quarter in LONG_QUARTER_PREFIXES:
            return LONG_QUARTER_PREFIXES[quarter]
        else:
            raise ValueError(
                "*** ERROR in lookup_epochs: Quarter must be between 0 and 17."
            )
    elif cadence == "short":
        if quarter in SHORT_QUARTER_PREFIXES:
            return SHORT_QUARTER_PREFIXES[quarter]
        else:
            raise ValueError(
                "*** ERROR in lookup_epochs: Quarter must be between 0 and 17."
            )
    else:
        raise ValueError(
            "*** ERROR in lookup_epochs: Cadence must be 'long' or 'short'."
        )


def print_cmd(ofile, url, cadence_str, kepid, epochs, cmdtype):
    # Print a data download command to an opened file (LUN) given a Kepler ID and array of epochs.
    for epoch in epochs:
        this_filename = "kplr" + kepid + "-" + epoch + cadence_str
        if cmdtype == "curl":
            cmd_str_to_add = " -f -R -o " + this_filename + " "
        else:
            cmd_str_to_add = " "
        this_cmd = (
            cmdtype
            + cmd_str_to_add
            + url
            + kepid[0:4]
            + "/"
            + kepid
            + "/"
            + this_filename
        )
        ofile.write(this_cmd + "\n")


def get_kepler(idict):
    # Extract command-line arguments from the input dict.
    kepids = idict["kepids"]
    cadence = idict["cadence"]
    data_type = idict["data_type"]
    ofile = idict["ofile"]
    cmdtype = idict["cmdtype"]
    if "epochs" in idict:
        epochs = idict["epochs"]
    else:
        epochs = None
    if "quarters" in idict:
        quarters = idict["quarters"]
    else:
        quarters = None

    # This is the base URL to use when requesting Kepler files.
    lc_base_url = "http://archive.stsci.edu/missions/kepler/lightcurves/"
    tp_base_url = "http://archive.stsci.edu/missions/kepler/target_pixel_files/"
    if data_type == "lightcurve":
        base_url = lc_base_url
    else:
        base_url = tp_base_url

    # Define the file extensions to use.
    if data_type == "lightcurve":
        if cadence == "long":
            cadence_str = "_llc.fits"
        else:
            cadence_str = "_slc.fits"
    else:
        if cadence == "long":
            cadence_str = "_lpd-targ.fits.gz"
        else:
            cadence_str = "_spd-targ.fits.gz"

    # Use all available quarters if neither 'epochs' or 'quarters' are specified.
    if epochs is None and quarters is None:
        quarters = [str(x) for x in range(18)]

    # If 'kepids' is a scalar string put it in a list.  Same with 'epochs' and 'quarters'
    if isinstance(kepids, str):
        kepids = [kepids]
    if isinstance(epochs, str):
        epochs = [epochs]
    if isinstance(quarters, str):
        quarters = [quarters]

    # Make sure Kepler IDs are zero-padded.
    kepids = ["{0:09d}".format(int(x)) for x in kepids]

    # Write to script file.
    with open(ofile, "w") as output_file:
        output_file.write("#!/bin/sh\n")
        for kepid in kepids:
            if quarters is not None:
                for quarter in quarters:
                    epoch_strings = lookup_epochs(quarter, cadence)
                    print_cmd(
                        output_file,
                        base_url,
                        cadence_str,
                        kepid,
                        epoch_strings,
                        cmdtype,
                    )
            else:
                print_cmd(output_file, base_url, cadence_str, kepid, epochs, cmdtype)
        # Print a warning message to let users know there will sometimes be Error 404's.
        output_file.write("echo\n")
        output_file.write(
            'echo Script completed.  Note: Some \\"Error 404\\" messages are expected,'
            " depending on your parameters.  It is always recommended to confirm the"
            " expected number of files are successfully downloaded.\n"
        )

    # Make sure file is executable.
    os.chmod(ofile, 0o744)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Script to generate download commands given one or"
        " more Kepler IDs.",
        epilog="Example: python get_kepler.py"
        " 7730747 7748238 -c short -t lightcurve -q 7 8",
    )
    parser.add_argument(
        "kepids",
        action="store",
        nargs="+",
        help="One or more Kepler IDs to" " retrieve data from.",
    )
    parser.add_argument(
        "-c",
        action="store",
        type=str.lower,
        dest="cadence",
        default="long",
        choices=["short", "long"],
        help="Specify the type of cadence.  Default=" "'%(default)s'.",
    )
    parser.add_argument(
        "-t",
        action="store",
        type=str.lower,
        dest="data_type",
        default="lightcurve",
        choices=["lightcurve", "target_pixel_file"],
        help="What type" " of data product to retrieve.  Default = '%(default)s'.",
    )
    parser.add_argument(
        "-o",
        action="store",
        dest="ofile",
        default="get_kepler_data.sh",
        help="Full"
        " path and file name of the output file to create.  Any existing file with"
        " the same name will be overwritten.  Default = '%(default)s'.",
    )
    parser.add_argument(
        "--cmdtype",
        action="store",
        type=str.lower,
        dest="cmdtype",
        default="curl",
        choices=["curl", "wget"],
        help="What download tool should be used?  Default" " = '%(default)s'.",
    )
    epoch_quarter_group = parser.add_mutually_exclusive_group()
    epoch_quarter_group.add_argument(
        "-e",
        action="store",
        type=str,
        dest="epochs",
        nargs="*",
        help="One or more epochs to retrieve with each Kepler ID.",
    )
    epoch_quarter_group.add_argument(
        "-q",
        action="store",
        type=str,
        dest="quarters",
        nargs="*",
        help="One or more quarters to retrieve with each Kepler ID.",
    )
    args = vars(parser.parse_args())
    get_kepler(args)



__all__ = [
    "get_transit_depth",
    "get_sma",
    "get_dur_14",
    "get_dur_23",
    "get_dur_cc",
    "predict_tc_error",
    "make_transit_mask",
    "set_oversample_factor",
]


import numpy as np
from .constants import *


def get_transit_depth(p, b):
    """
    Calculate approximate transit depth
    See Mandel & Agol 2002

    Parameters
    ----------
    p : planet-to-star radius ratio Rp/Rstar
    b : impact parameter

    Returns
    -------
    d : transit depth
    """
    # broadcasting
    p = p * np.ones(np.atleast_1d(b).shape)
    b = b * np.ones(np.atleast_1d(p).shape)

    # non-grazing transit (b <= 1-p)
    d = p**2

    # grazing transit (1-p < b < 1+p)
    grazing = (b > 1 - p) * (b < 1 + p)

    pg = p[grazing]
    bg = b[grazing]

    k0 = np.arccos((pg**2 + bg**2 - 1) / (2 * pg * bg))
    k1 = np.arccos((1 - pg**2 + bg**2) / (2 * bg))
    s0 = np.sqrt((4 * bg**2 - (1 + bg**2 - pg**2) ** 2) / 4)

    d[grazing] = (1 / pi) * (pg**2 * k0 + k1 - s0)

    # non-transiting (b >= 1+p)
    d[b >= 1 + p] = 0.0

    return np.squeeze(d)


def get_sma(P, Ms):
    """
    Calculate semi-major axis in units of Solar radii from Kepler's law

    Parameters
    ----------
    P : period [days]
    Ms : stellar mass [Solar masses]

    Returns
    -------
    sma : semimajor axis [Solar radii]
    """
    return Ms ** (1 / 3) * (P / 365.24) ** (2 / 3) / RSAU


def get_dur_14(P, aRs, b, ror, ecc=None, w=None):
    """
    Calculate total transit duration (I-IV contacts)
    See Winn 2010 Eq. 14 & 16

    P : period
    aRs : semimajor axis over stellar radius
    b : impact parameter
    ror : planet-to-star radius ratio
    ecc : eccentricity
    w : argument of periastron [radians]
    """
    if ecc is not None:
        Xe = np.sqrt(1 - ecc**2) / (1 + ecc * np.sin(w))
        We = (1 - ecc**2) / (1 + ecc * np.sin(w))
    else:
        Xe = 1.0
        We = 1.0

    sini = np.sin(np.arccos((b / We) / (aRs)))
    argument = (1 / aRs) * np.sqrt((1 + ror) ** 2 - b**2) / sini

    Ttot = (P / pi) * np.arcsin(argument) * Xe

    return Ttot


def get_dur_23(P, aRs, b, ror, ecc=None, w=None):
    """
    Calculate full duration (II-III contacts)
    See Winn 2010 Eq. 15 & 16

    P : period
    aRs : semimajor axis over stellar radius
    b : impact parameter
    ror : planet-to-star radius ratio
    ecc : eccentricity
    w : argument of periastron [radians]
    """
    if ecc is not None:
        Xe = np.sqrt(1 - ecc**2) / (1 + ecc * np.sin(w))
        We = (1 - ecc**2) / (1 + ecc * np.sin(w))
    else:
        Xe = 1.0
        We = 1.0

    sini = np.sin(np.arccos((b / We) / (aRs)))
    argument = (1 / aRs) * np.sqrt((1 - ror) ** 2 - b**2) / sini

    Tfull = np.asarray((P / pi) * np.arcsin(argument)) * Xe

    # correct for grazing transits
    grazing = np.asarray(b) > np.asarray(1 - ror)
    Tfull[grazing] = np.nan

    return Tfull


def get_dur_cc(P, aRs, b, ecc=None, w=None):
    """
    Calculate ingress/egrees midpoint transit duration (1.5-3.5 contacts)
    See Winn 2010

    P : period
    aRs : semimajor axis over stellar radius
    b : impact parameter
    ecc : eccentricity
    w : argument of periastron [radians]
    """
    if ecc is not None:
        Xe = np.sqrt(1 - ecc**2) / (1 + ecc * np.sin(w))
        We = (1 - ecc**2) / (1 + ecc * np.sin(w))
    else:
        Xe = 1.0
        We = 1.0

    sini = np.sin(np.arccos((b / We) / (aRs)))
    argument = (1 / aRs) * np.sqrt(1 - b**2) / sini

    Tmid = np.asarray((P / pi) * np.arcsin(argument)) * Xe

    # correct for grazing transits
    grazing = np.asarray(b) > 1
    Tmid[grazing] = np.nan

    return Tmid


def predict_tc_error(ror, b, T14, texp, sigma_f):
    """
    Predict uncertainty on mid-transit time
    See Carter+2008

    Parameters
    ----------
    ror : planet-to-star radius ratio
    b : impact parameter
    T14 : first-to-fourth contact transit duration
    texp : exposure time (in same units as T14)
    sigma_f : photometric error on flux
    """
    Q = np.sqrt(T14 / texp) * (ror**2 / sigma_f)

    if b <= 1 - ror:
        tau_over_T14 = ror / (1 - b**2)
    else:
        tau_over_T14 = 0.5

    sigma_tc = (T14 / Q) * np.sqrt(0.5 * tau_over_T14)

    return sigma_tc


def make_transit_mask(time, tts, masksize):
    """
    Make a transit mask for an alderaan.Planet() object

    Parameters
    ----------
        time : array-like
            time values at each cadence
        tts : array-like
            transit times for a single planet
        masksize : float
            size of mask window in units of time

    Returns
    -------
        transitmask : ndarray, bool
            boolean array (1=near transit; 0=not)
    """
    transitmask = np.zeros(len(time), dtype="bool")

    tts_here = tts[(tts >= time.min()) * (tts <= time.max())]

    for t0 in tts_here:
        neartransit = np.abs(time - t0) < masksize
        transitmask += neartransit

    return transitmask


def set_oversample_factor(periods, depths, durs, flux, error, texp):
    """
    Docstring
    """
    if (len(periods) != len(depths)) or (len(periods) != len(durs)):
        raise ValueError("Input array shape mismatch")

    npl = len(periods)

    # ingress/egress timescale estimate following Winn 2010
    ror = np.sqrt(depths)
    tau = 13 * (periods / 365.25) ** (1 / 3) * ror / 24

    # set sigma so binning error is < 0.1% of photometric uncertainty
    sigma = np.mean(error / flux) * 0.04

    npts = np.array(np.ceil(np.sqrt((depths / tau) * (texp / 8 / sigma))), dtype="int")
    npts = npts + (npts % 2 + 1)

    npts = np.max(np.hstack([npts, 7]))

    return npts



__all__ = ['bin_data',
           'estimate_transit_depth',
           ]


import numpy as np
from alderaan.constants import pi


def bin_data(time, data, binsize, bin_centers=None):
    """Bin a 1D timeseries data array

    If bin_centers is not provided, bins will be linearly spaced
    between (time.min(),time.max())

    Args:
        time (ndarray) : time values
        data (ndarray) : corresponding data to be binned
        binsize (float): bin size for output data, in same units as time
        bin_centers (optional) : pre-defined bin centers

    Returns:
        tuple : a tuple containing:
          - time_binned (array-like)
          - data_binned (array-like)
    """
    if bin_centers is None:
        bin_centers = np.hstack(
            [
                np.arange(time.mean(), time.min() - binsize / 2, -binsize)[::-1],
                np.arange(time.mean(), time.max() + binsize / 2, binsize)[1:],
            ]
        )

    binned_data = np.zeros(len(bin_centers))
    for i, t0 in enumerate(bin_centers):
        binned_data[i] = np.mean(data[np.abs(time - t0) < binsize / 2])

    return bin_centers, binned_data


def estimate_transit_depth(p, b):
    """Calculate approximate transit depth following Mandel & Agol (2002)

    Args:
        p (array-like) : planet-to-star radius ratio Rp/Rstar
        b (array-like) : impact parameter

    Returns:
        array-like : transit depth
    """
    # broadcasting
    p = p * np.ones(np.atleast_1d(b).shape)
    b = b * np.ones(np.atleast_1d(p).shape)

    # non-grazing transit (b <= 1-p)
    d = p**2

    # grazing transit (1-p < b < 1+p)
    grazing = (b > 1 - p) * (b < 1 + p)

    pg = p[grazing]
    bg = b[grazing]

    k0 = np.arccos((pg**2 + bg**2 - 1) / (2 * pg * bg))
    k1 = np.arccos((1 - pg**2 + bg**2) / (2 * bg))
    s0 = np.sqrt((4 * bg**2 - (1 + bg**2 - pg**2) ** 2) / 4)

    d[grazing] = (1 / pi) * (pg**2 * k0 + k1 - s0)

    # non-transiting (b >= 1+p)
    d[b >= 1 + p] = 0.0

    return np.squeeze(d)


__all__ = ['resolve_config_path',
           'parse_koi_catalog',
           'parse_holczer16_catalog',
           'copy_input_target_catalog',
          ]


import os
import sys
from pathlib import Path

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from alderaan.ephemeris import Ephemeris


def resolve_config_path(path_str, base_path):
    return os.path.join(str(Path(path_str.format(base_path=base_path)).resolve()),'')


def parse_koi_catalog(filepath, koi_id):
    """Reads a Kepler Object of Interest Catalog and performs consistency checks
    
    Expected columns are:
        [kic_id, koi_id, npl, period, epoch, depth, duration, impact, ld_u1, ld_u2]

    Args:
        filepath (str) : path to csv file
        koi_id (str) : KOI identification number in the format, e.g., K01234

    Returns:
        pd.DataFrame: catalog of target star/planet properties
    """
    # read catalog from csv file
    catalog = pd.read_csv(filepath, index_col=0)
    catalog = catalog.loc[catalog.koi_id == koi_id]

    # sort by ascending period
    catalog = catalog.sort_values(by='period').reset_index(drop=True)

    # check for consistency in multi-planet systems
    if not all(kic_id == catalog.kic_id.to_numpy()[0] for kic_id in catalog.kic_id):
        raise ValueError("There are inconsistencies with KIC in the csv input file")

    if not all(npl == catalog.npl.to_numpy()[0] for npl in catalog.npl):
        raise ValueError("There are inconsistencies with NPL in the csv input file")

    if not all(
        ld_u1 == catalog.limbdark_1.to_numpy()[0] for ld_u1 in catalog.limbdark_1
    ):
        raise ValueError("There are inconsistencies with LD_U1 in the csv input file")

    if not all(
        ld_u2 == catalog.limbdark_2.to_numpy()[0] for ld_u2 in catalog.limbdark_2
    ):
        raise ValueError("There are inconsistencies with LD_U2 in the csv input file")

    # check for NaN valued transit parameters
    if np.any(
        np.isnan(np.array(catalog["period epoch depth duration impact".split()]))
    ):
        raise ValueError("NaN values found in input catalog")
    
    return catalog


def parse_holczer16_catalog(filepath, koi_id, num_planets):
    """Reads transit time table from Holczer+2016 into a list of Ephemeris objects

    Automatically corrects for zero-point offsets between catalogs
      - Holczer+2016 used BJD - 2454900
      - Kepler Project used BJKD = BJD - 2454833

    Args:
        filepath (str) : path to Holczer+2016 table
        koi_id (str) : KOI identification number in the format, e.g., K01234
        num_planets : total number of planets in the system

    Returns:
        list : list of (0,num_planets) Ephemeris objects
    """
    data = np.loadtxt(filepath, usecols=[0,1,2,3,4], dtype=str)
    ephemerides = []

    planet_id = data[:,0]
    index = np.array(data[:,1], dtype=int)
    ttime = np.array(data[:,2], dtype=float) + np.array(data[:,3],dtype=float)/24/60 + 67
    error = np.array(data[:,4], dtype=float)/24/60

    for n in range(num_planets):
        use = planet_id == f"{int(koi_id[1:])}.0{1+n}"
        if np.sum(use) > 0:
            ephemerides.append(Ephemeris(index=index[use], ttime=ttime[use], error=error[use]))

    return ephemerides


def copy_input_target_catalog(filepath_master, filepath_copy):
    df_master = pd.read_csv(filepath_master, index_col=0)
    
    if os.path.exists(filepath_copy):
        df_copy = pd.read_csv(filepath_copy, index_col=0)

        try:
            assert_frame_equal(df_master, df_copy, check_like=True)
        except AssertionError:
            print(f"AssertionError: existing file {filepath_copy} does not match active file {filepath_master}")

    else:
        os.makedirs(os.path.dirname(filepath_copy), exist_ok=True)
        df_master.to_csv(filepath_copy)



__all__ = [
    "plot_omc",
    "plot_holczer",
    "plot_omc_model_selection",
    "plot_folded_transit",
    "plot_tc_vs_chisq",
    "plot_acf",
    "plot_synthetic_noise",
]


from astropy.stats import mad_std
import numpy as np
import matplotlib.pyplot as plt
from .utils import bin_data


def plot_omc(linear_ephemeris, tts_obs, tts_mod, out=None, colors="auto"):
    """
    Docstring
    """
    npl = len(linear_ephemeris)

    if out is None:
        out = []
        for n in range(npl):
            out.append(np.zeros(len(linear_ephemeris[n]), dtype="bool"))

    if colors == "auto":
        colors = []
        for n in range(npl):
            colors.append(f"C{n}")

    fig, ax = plt.subplots(npl, figsize=(8, 2 * npl))
    if npl == 1:
        ax = [ax]

    for n in range(npl):
        xtime = linear_ephemeris[n]
        yomc_obs = (tts_obs[n] - linear_ephemeris[n]) * 24 * 60
        yomc_mod = (tts_mod[n] - linear_ephemeris[n]) * 24 * 60

        ax[n].plot(xtime, yomc_obs, "o", c="lightgrey")
        ax[n].plot(xtime, yomc_mod, lw=2, c=colors[n])
        ax[n].plot(xtime[out[n]], yomc_obs[out[n]], "rx")
        ax[n].set_ylabel("O-C [min]", fontsize=14)
    ax[npl - 1].set_xlabel("Time [BJKD]", fontsize=14)

    return fig, ax


def plot_holczer(data, n):
    xtime = data["tts"][n]
    yomc = data["tts"][n] - data["epoch"][n] - data["period"][n] * data["inds"][n]
    out = data["out"][n]

    xtime_interp = data["full_tts"][n]
    yomc_interp = (
        data["full_tts"][n]
        - data["epoch"][n]
        - data["period"][n] * data["full_inds"][n]
    )

    fig = plt.figure(figsize=(8, 3))

    plt.plot(xtime[~out], yomc[~out] * 24 * 60, "o", c="grey", label="Holczer+2016")
    plt.plot(xtime[out], yomc[out] * 24 * 60, "rx")
    plt.plot(xtime_interp, yomc_interp * 24 * 60, "k+", label="Interpolation")
    plt.xlabel("Time [BJKD]", fontsize=16)
    plt.ylabel("O-C [min]", fontsize=16)
    plt.legend(fontsize=12)

    return fig


def plot_tc_vs_chisq(t_obs, f_obs, t_mod, f_mod, tc_fit, x2_fit, x2_mod, tc, color):
    """
    Docstring
    """
    fig, ax = plt.subplots(1, 2, figsize=(8, 3))

    ax[0].plot((t_obs - tc) * 24, (f_obs - 1) * 1e6, "ko")
    ax[0].plot(t_mod * 24, (f_mod - 1) * 1e6, c=color, lw=2)
    ax[0].set_xlabel(r"$\Delta t$ (hours)", fontsize=14)
    ax[0].set_ylabel(r"$\Delta F$ (ppm)", fontsize=14)

    ax[1].plot((tc_fit - tc) * 24, x2_fit - x2_fit.min(), "ko")
    ax[1].plot((tc_fit - tc) * 24, x2_mod - x2_fit.min(), c=color, lw=3)
    ax[1].axvline(0, color="k", ls=":", lw=2)
    ax[1].set_xlabel(r"$\Delta t$ (hours)", fontsize=14)
    ax[1].set_ylabel(r"$\Delta \chi^2$", fontsize=14)

    plt.tight_layout()

    return fig, ax


def plot_omc_model_selection(
    xtime, yomc, fg_prob, bad, full_linear_ephemeris, full_omc_trend, err, rms, target
):
    """
    Docstring
    """
    fig = plt.figure(figsize=(12, 4))

    plt.scatter(xtime, yomc * 24 * 60, c=1 - fg_prob, cmap="viridis", label="MAP TTVs")
    plt.plot(xtime[bad], yomc[bad] * 24 * 60, "rx")
    plt.plot(
        full_linear_ephemeris, full_omc_trend * 24 * 60, "k", label="Regularized model"
    )
    plt.xlabel("Time [BJKD]", fontsize=20)
    plt.ylabel("O-C [min]", fontsize=20)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.legend(fontsize=14, loc="upper right")
    plt.title(target, fontsize=20)
    plt.text(
        xtime.min(),
        yomc.min() * 24 * 60,
        f"measured error = {err:.1f} min",
        fontsize=16,
        ha="left",
        backgroundcolor="w",
    )
    plt.text(
        xtime.max(),
        yomc.min() * 24 * 60,
        f"residual RMS = {rms:.1f} min",
        fontsize=16,
        ha="right",
        backgroundcolor="w",
    )

    return fig


def plot_folded_transit(lc, sc, tts, depth, duration, target, n):
    """
    Docstring
    """
    t_folded = []
    f_folded = []

    # grab the data
    for t0 in tts:
        if sc is not None:
            use = np.abs(sc.time - t0) / duration < 1.5

            if np.sum(use) > 0:
                t_folded.append(sc.time[use] - t0)
                f_folded.append(sc.flux[use])

        if lc is not None:
            use = np.abs(lc.time - t0) / duration < 1.5

            if np.sum(use) > 0:
                t_folded.append(lc.time[use] - t0)
                f_folded.append(lc.flux[use])

    # sort the data
    t_folded = np.hstack(t_folded)
    f_folded = np.hstack(f_folded)

    order = np.argsort(t_folded)
    t_folded = t_folded[order]
    f_folded = f_folded[order]

    # bin the data
    t_binned, f_binned = bin_data(t_folded, f_folded, duration / 11)

    # set undersampling factor and plotting limits
    inds = np.arange(len(t_folded), dtype="int")
    inds = np.random.choice(inds, size=np.min([3000, len(inds)]), replace=False)

    ymin = -3 * mad_std(f_folded) - depth
    ymax = +3 * mad_std(f_folded)

    # plot the data
    fig = plt.figure(figsize=(8, 3))
    plt.plot(t_folded[inds] * 24, (f_folded[inds] - 1) * 1e6, ".", c="lightgrey")
    plt.plot(
        t_binned * 24,
        (f_binned - 1) * 1e6,
        "o",
        ms=8,
        color=f"C{n}",
        label=f"{target}-{n}",
    )
    plt.xlim(t_folded.min() * 24, t_folded.max() * 24)
    plt.ylim(ymin * 1e6, ymax * 1e6)
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)
    plt.xlabel("Time from mid-transit [hrs]", fontsize=16)
    plt.ylabel("Flux (ppm)", fontsize=16)
    plt.legend(fontsize=12, loc="lower right", framealpha=1)

    return fig


def plot_acf(xcor, acf_emp, acf_mod, xf, yf, freqs, target_name, season):
    """
    Docstring
    """
    fig = plt.figure(figsize=(20, 5))

    plt.tight_layout()
    plt.subplots_adjust(wspace=0.8)

    ax = plt.subplot2grid(shape=(5, 10), loc=(0, 0), rowspan=3, colspan=7)
    ax.plot(xcor * 24, acf_emp, color="lightgrey")
    ax.plot(xcor * 24, acf_mod, c="red")
    ax.set_xlim(xcor.min() * 24, xcor.max() * 24)
    ax.set_xticks(np.arange(0, xcor.max() * 24, 2))
    ax.set_xticklabels([])
    ax.set_ylim(acf_emp.min() * 1.1, acf_emp.max() * 1.1)
    ax.set_ylabel("ACF", fontsize=20)
    ax.text(
        xcor.max() * 24 - 0.15,
        acf_emp.max(),
        f"{target_name}, {season}",
        va="top",
        ha="right",
        fontsize=20,
    )

    ax = plt.subplot2grid(shape=(5, 10), loc=(0, 7), rowspan=5, colspan=3)
    ax.plot(xf / 24 / 3600 * 1e3, yf, color="k", lw=0.5)
    for f in freqs:
        ax.axvline(f / 24 / 3600 * 1e3, color="red", zorder=0, lw=3, alpha=0.3)
    ax.set_xlim(xf.min() / 24 / 3600 * 1e3, xf.max() / 24 / 3600 * 1e3)
    ax.set_ylim(yf.min(), 1.2 * yf.max())
    ax.set_ylabel("Power", fontsize=20)
    ax.set_yticks([])
    ax.set_xlabel("Frequency [mHz]", fontsize=20)

    for i, sf in enumerate(np.sort(freqs)[::-1]):
        ax.text(
            xf.min() / 24 / 3600 * 1e3 + 0.1,
            yf.max() * (1.1 - 0.1 * i),
            f"{0:.2f} min".format(24 * 60 / sf),
            fontsize=16,
        )

    ax = plt.subplot2grid(shape=(5, 10), loc=(3, 0), rowspan=2, colspan=7)
    ax.plot(xcor * 24, acf_emp - acf_mod, c="lightgrey")
    ax.set_xlim(xcor.min() * 24, xcor.max() * 24)
    ax.set_xticks(np.arange(0, xcor.max() * 24, 2))
    ax.set_xlabel("Lag time [hours]", fontsize=20)
    ax.set_ylabel("Residuals", fontsize=20)

    return fig


def plot_synthetic_noise(t, white_noise, red_noise, target, season, depths):
    """
    Docstring
    """
    fig = plt.figure(figsize=(20, 5))

    plt.plot(t, white_noise + red_noise, ".", c="lightgrey")
    plt.plot(t, red_noise, c="r", lw=4, label=f"{target}, SEASON {season}")
    plt.axhline(depths.max(), c="k", ls=":", lw=2)
    plt.axhline(depths.min(), c="k", ls="--", lw=2)
    plt.axhline(-depths.min(), c="k", ls="--", lw=2)
    plt.axhline(-depths.max(), c="k", ls=":", lw=2)
    plt.xlim(t.min(), t.max())
    plt.ylim(np.percentile(white_noise, 1), np.percentile(white_noise, 99))
    plt.xlabel("Time [days]", fontsize=24)
    plt.ylabel("Flux", fontsize=24)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.legend(fontsize=20, loc="upper right", framealpha=1)
    # plt.savefig(os.path.join(FIGURE_DIR, TARGET + f'_synthetic_noise_season_{z}.png'), bbox_inches='tight')

    return fig



__all__ = ["psi_N", "psi_T", "psi_G", "F_iter", "z_iter", "emus_weights"]


import numpy as np
import scipy.linalg as linalg


def psi_N(x, norm=1.0):
    """
    Bias funciton for 'N' (non-grazing) umbrella; see Gilbert 2022
    """
    x_ = np.atleast_1d(x)
    psi = np.zeros_like(x_)
    psi[(x_ < 2) * (x_ >= 1)] = x_[(x_ < 2) * (x_ >= 1)] - 1
    psi[(x_ >= 2)] = 1.0
    psi = psi.clip(0, 1)

    return psi / norm


def psi_T(x, norm=1.0):
    """
    Bias funciton for 'T' (transition) umbrella; see Gilbert 2022
    """
    x_ = np.atleast_1d(x)
    psi = np.zeros_like(x_)
    psi[(x_ >= 0) * (x_ < 1)] = x_[(x_ >= 0) * (x_ < 1)]
    psi[(x_ >= 1) * (x_ < 2)] = 2 - x_[(x_ >= 1) * (x_ < 2)]
    psi = psi.clip(0, 1)

    return psi / norm


def psi_G(x, norm=1.0):
    """
    Bias funciton for 'G' (grazing) umbrella; see Gilbert 2022
    """
    x_ = np.atleast_1d(x)
    psi = np.zeros_like(x_)
    psi[(x_ < 0) * (x_ >= -1)] = 1 + x_[(x_ < 0) * (x_ >= -1)]
    psi[(x_ >= 0)] = 1.0 - x_[x_ >= 0]
    psi = psi.clip(0, 1)

    return psi / norm


def F_iter(z, psi_fxns, coordinates, weights=None):
    """
    Helper function for .emus_weights
    """
    Nwin = len(psi_fxns)
    F = np.zeros((Nwin, Nwin))

    if weights is None:
        weights = []
        for j, c_ in enumerate(coordinates):
            weights.append(np.ones(len(c_)) / len(c_))

    for i in range(Nwin):
        denom = 0.0
        for k in range(Nwin):
            denom += psi_fxns[k](coordinates[i]) / z[k]
        for j in range(Nwin):
            num = psi_fxns[j](coordinates[i]) / z[i]
            F[i, j] = np.sum(weights[i] * num / denom)

    return F


def z_iter(F, tol=1.0e-10, max_iter=100):
    """
    Helper function for emus_weights
    """
    # stationary distribution is the last column of QR factorization
    M = np.eye(len(F)) - F
    q, r = linalg.qr(M)
    z = q[:, -1]
    z /= np.sum(z)

    # polish solution using power method.
    for itr in range(max_iter):
        znew = np.dot(z, F)
        tv = np.abs(znew[z > 0] - z[z > 0])
        tv = tv / z[z > 0]

        maxresid = np.max(tv)
        if maxresid < tol:
            break
        else:
            z = znew

    # return normalized (by convention)
    return z / np.sum(z)


def emus_weights(
    psi_fxns, coordinates, weights=None, nMBAR=20, tol=1e-10, max_iter=100
):
    """
    Calculate umbrella weights using the EMUS algorithm
    See Matthews+ 2017 (https://ui.adsabs.harvard.edu/abs/2018MNRAS.480.4069M/abstract)

    Parameters
    ----------
    psi_fxns : list
        list of callable 1D functions, each takes a 1D vector of coordinates and returns normalized psi
    coordinates : list
        list of 1D arrays, each array is the coordinates for samples from a single umbrella
    weights : list
        list of 1D arrays, each array is sample weights corresponding to coordinates (default = None)
    nMBAR : int
        number of MBAR iterations to perform
    tol : float
        convergence tolerance
    max_iter
        maximum number of polishing iterations to perform on z

    Returns
    -------
    z : ndarray
        normalized window weights
    """
    Nwin = len(psi_fxns)

    z = np.zeros(Nwin)

    for i, psi in enumerate(psi_fxns):
        if weights is None:
            z[i] = np.mean(psi(coordinates[i]))
        else:
            if np.abs(np.sum(weights[i]) - 1) > 1e-8:
                raise ValueError("sample weights must sum to unity")
            else:
                z[i] = np.sum(psi(coordinates[i]) * weights[i])

    z /= np.sum(z)

    for n in range(nMBAR):
        F = F_iter(z, psi_fxns, coordinates, weights)
        z = z_iter(F)

        # perform self-consistent polishing until convergence
        for n in range(nMBAR):
            z_old = np.copy(z)
            z_old[z_old < 1e-100] = 1e-100

            F = F_iter(z, psi_fxns, coordinates, weights)
            z = z_iter(F, tol, max_iter)

            # check if we have converged
            if np.max(np.abs(z - z_old) / z_old) < tol:
                break

    return z



# Yeah, there's not much here
# I've migrated features elsewhere, so this is now a glorifed dictionary

__all__ = ["Planet"]


class Planet:
    def __init__(
        self,
        period=None,
        epoch=None,
        depth=None,
        duration=None,
        impact=None,
        tts=None,
        index=None,
        quality=None,
        overlap=None,
        dtype=None,
    ):
        self.period = period  # orbital period
        self.epoch = epoch  # reference transit time in range (0, period)
        self.depth = depth  # transit depth
        self.duration = duration  # transit duration
        self.impact = impact  # impact parameter

        self.tts = tts  # midtransit times
        self.index = index  # index of each transit time
        self.quality = quality  # bool flag per transit; True=good
        self.overlap = overlap  # bool flag per transit; True = transit overlaps with another planet
